from .base_config import BaseConfig
from .config import ColBERTConfig, RunConfig
from .settings import (
    DocSettings,
    IndexingSettings,
    ResourceSettings,
    RunSettings,
    SearchSettings,
)

__all__ = [
    "RunSettings",
    "TokenizerSettings",
    "ResourceSettings",
    "DocSettings",
    "IndexingSettings",
    "SearchSettings",
    "ColBERTConfig",
    "RunConfig",
    "BaseConfig",
]



import copy
from argparse import ArgumentParser

# import faiss
import pylate.indexes.stanford_nlp.utils.distributed as distributed
from pylate.indexes.stanford_nlp.utils.runs import Run


class Arguments:
    def __init__(self, description):
        self.parser = ArgumentParser(description=description)
        self.checks = []

        self.add_argument("--root", dest="root", default="experiments")
        self.add_argument("--experiment", dest="experiment", default="dirty")
        self.add_argument("--run", dest="run", default=Run.name)

        self.add_argument("--local_rank", dest="rank", default=-1, type=int)

    def add_model_parameters(self):
        # Core Arguments
        self.add_argument(
            "--similarity",
            dest="similarity",
            default="cosine",
            choices=["cosine", "l2"],
        )
        self.add_argument("--dim", dest="dim", default=128, type=int)
        self.add_argument("--query_maxlen", dest="query_maxlen", default=32, type=int)
        self.add_argument("--doc_maxlen", dest="doc_maxlen", default=180, type=int)

        # Filtering-related Arguments
        self.add_argument(
            "--mask-punctuation",
            dest="mask_punctuation",
            default=False,
            action="store_true",
        )

    def add_model_training_parameters(self):
        # NOTE: Providing a checkpoint is one thing, --resume is another, --resume_optimizer is yet another.
        self.add_argument("--resume", dest="resume", default=False, action="store_true")
        self.add_argument(
            "--resume_optimizer",
            dest="resume_optimizer",
            default=False,
            action="store_true",
        )
        self.add_argument(
            "--checkpoint", dest="checkpoint", default=None, required=False
        )

        self.add_argument("--lr", dest="lr", default=3e-06, type=float)
        self.add_argument("--maxsteps", dest="maxsteps", default=400000, type=int)
        self.add_argument("--bsize", dest="bsize", default=32, type=int)
        self.add_argument("--accum", dest="accumsteps", default=2, type=int)
        self.add_argument("--amp", dest="amp", default=False, action="store_true")

    def add_model_inference_parameters(self):
        self.add_argument("--checkpoint", dest="checkpoint", required=True)
        self.add_argument("--bsize", dest="bsize", default=128, type=int)
        self.add_argument("--amp", dest="amp", default=False, action="store_true")

    def add_training_input(self):
        self.add_argument("--triples", dest="triples", required=True)
        self.add_argument("--queries", dest="queries", default=None)
        self.add_argument("--collection", dest="collection", default=None)

        def check_training_input(args):
            assert (args.collection is None) == (args.queries is None), (
                "For training, both (or neither) --collection and --queries must be supplied."
                "If neither is supplied, the --triples file must contain texts (not PIDs)."
            )

        self.checks.append(check_training_input)

    def add_ranking_input(self):
        self.add_argument("--queries", dest="queries", default=None)
        self.add_argument("--collection", dest="collection", default=None)
        self.add_argument("--qrels", dest="qrels", default=None)

    def add_reranking_input(self):
        self.add_ranking_input()
        self.add_argument("--topk", dest="topK", required=True)
        self.add_argument(
            "--shortcircuit", dest="shortcircuit", default=False, action="store_true"
        )

    def add_indexing_input(self):
        self.add_argument("--collection", dest="collection", required=True)
        self.add_argument("--index_root", dest="index_root", required=True)
        self.add_argument("--index_name", dest="index_name", required=True)

    def add_compressed_index_input(self):
        self.add_argument(
            "--compression_level",
            dest="compression_level",
            choices=[1, 2],
            type=int,
            default=None,
        )

    def add_index_use_input(self):
        self.add_argument("--index_root", dest="index_root", required=True)
        self.add_argument("--index_name", dest="index_name", required=True)
        self.add_argument(
            "--partitions", dest="partitions", default=None, type=int, required=False
        )

    def add_retrieval_input(self):
        self.add_index_use_input()
        self.add_argument("--nprobe", dest="nprobe", default=10, type=int)
        self.add_argument(
            "--retrieve_only", dest="retrieve_only", default=False, action="store_true"
        )

    def add_argument(self, *args, **kw_args):
        return self.parser.add_argument(*args, **kw_args)

    def check_arguments(self, args):
        for check in self.checks:
            check(args)

    def parse(self):
        args = self.parser.parse_args()
        self.check_arguments(args)

        args.input_arguments = copy.deepcopy(args)

        args.nranks, args.distributed = distributed.init(args.rank)

        # args.nthreads = int(max(os.cpu_count(), faiss.omp_get_max_threads()) * 0.8)
        args.nthreads = max(1, args.nthreads // args.nranks)

        # if args.nranks > 1:
        #     print_message(
        #         f"#> Restricting number of threads for FAISS to {args.nthreads} per process",
        #         condition=(args.rank == 0),
        #     )
        #     faiss.omp_set_num_threads(args.nthreads)

        Run.init(args.rank, args.root, args.experiment, args.run)
        Run._log_args(args)
        Run.info(args.input_arguments.__dict__, "\n")

        return args



from __future__ import annotations

import dataclasses
import os
from dataclasses import dataclass

import ujson

from .core_config import CoreConfig


@dataclass
class BaseConfig(CoreConfig):
    @classmethod
    def from_existing(cls, *sources):
        kw_args = {}

        for source in sources:
            if source is None:
                continue

            local_kw_args = dataclasses.asdict(source)
            local_kw_args = {k: local_kw_args[k] for k in source.assigned}
            kw_args = {**kw_args, **local_kw_args}

        obj = cls(**kw_args)

        return obj

    @classmethod
    def from_deprecated_args(cls, args):
        obj = cls()
        ignored = obj.configure(ignore_unrecognized=True, **args)

        return obj, ignored

    @classmethod
    def from_path(cls, name):
        with open(name) as f:
            args = ujson.load(f)

            if "config" in args:
                args = args["config"]

        return cls.from_deprecated_args(
            args
        )  # the new, non-deprecated version functions the same at this level.

    @classmethod
    def load_from_index(cls, index_path):
        # FIXME: We should start here with initial_config = ColBERTConfig(config, Run().config).
        # This should allow us to say initial_config.index_root. Then, below, set config = Config(..., initial_c)

        # default_index_root = os.path.join(Run().root, Run().experiment, 'indexes/')
        # index_path = os.path.join(default_index_root, index_path)

        # CONSIDER: No more plan/metadata.json. Only metadata.json to avoid weird issues when loading an index.

        try:
            metadata_path = os.path.join(index_path, "metadata.json")
            loaded_config, _ = cls.from_path(metadata_path)
        except Exception:
            metadata_path = os.path.join(index_path, "plan.json")
            loaded_config, _ = cls.from_path(metadata_path)

        return loaded_config

    def save(self, path, overwrite=False):
        assert overwrite or not os.path.exists(path), path

        with open(path, "w") as f:
            args = self.export()  # dict(self.__config)
            # TODO: Add git_status details.. It can't be too large! It should be a path that Runs() saves on exit, maybe!

            f.write(ujson.dumps(args, indent=4) + "\n")

    def save_for_checkpoint(self, checkpoint_path):
        assert not checkpoint_path.endswith(".dnn"), (
            f"{checkpoint_path}: We reserve *.dnn names for the deprecated checkpoint format."
        )

        output_config_path = os.path.join(checkpoint_path, "artifact.metadata")
        self.save(output_config_path, overwrite=True)



from __future__ import annotations

from .index_updater import IndexUpdater
from .indexer import Indexer
from .searcher import Searcher

__all__ = [
    "Indexer",
    "IndexUpdater",
    "Searcher",
]



import os
import re

import ujson


def load_doclens(directory, flatten=True):
    doclens_filenames = {}

    for filename in os.listdir(directory):
        match = re.match(r"doclens.(\d+).json", filename)

        if match is not None:
            doclens_filenames[int(match.group(1))] = filename

    doclens_filenames = [
        os.path.join(directory, doclens_filenames[i])
        for i in sorted(doclens_filenames.keys())
    ]

    all_doclens = []
    for filename in doclens_filenames:
        with open(filename) as f:
            all_doclens.append(ujson.load(f))

    if flatten:
        all_doclens = [x for sub_doclens in all_doclens for x in sub_doclens]

    if len(all_doclens) == 0:
        raise ValueError("Could not load doclens")

    return all_doclens



import os
import pathlib

from torch.utils.cpp_extension import load

from pylate.indexes.stanford_nlp.infra.config.config import ColBERTConfig
from pylate.indexes.stanford_nlp.search.strided_tensor import StridedTensor
from pylate.indexes.stanford_nlp.utils.utils import print_message


def try_load_torch_extensions(use_gpu):
    print_message(
        "Loading segmented_maxsim_cpp extension (set COLBERT_LOAD_TORCH_EXTENSION_VERBOSE=True for more info)..."
    )
    segmented_maxsim_cpp = load(
        name="segmented_maxsim_cpp",
        sources=[
            os.path.join(
                pathlib.Path(__file__).parent.resolve(), "segmented_maxsim.cpp"
            ),
        ],
        extra_cflags=["-O3"],
        verbose=os.getenv("COLBERT_LOAD_TORCH_EXTENSION_VERBOSE", "False") == "True",
    )

    return segmented_maxsim_cpp.segmented_maxsim_cpp


# TODO: The masking below might also be applicable in the kNN part
def colbert_score_reduce(scores_padded, D_mask, config: ColBERTConfig):
    D_padding = ~D_mask.view(scores_padded.size(0), scores_padded.size(1)).bool()
    scores_padded[D_padding] = -9999
    scores = scores_padded.max(1).values

    return scores.sum(-1)


# TODO: Wherever this is called, pass `config=`
def colbert_score(Q, D_padded, D_mask, config=ColBERTConfig()):
    """
    Supply sizes Q = (1 | num_docs, *, dim) and D = (num_docs, *, dim).
    If Q.size(0) is 1, the matrix will be compared with all passages.
    Otherwise, each query matrix will be compared against the *aligned* passage.

    EVENTUALLY: Consider masking with -inf for the maxsim (or enforcing a ReLU).
    """

    use_gpu = config.total_visible_gpus > 0
    if use_gpu:
        Q, D_padded, D_mask = Q.cuda(), D_padded.cuda(), D_mask.cuda()

    assert Q.dim() == 3, Q.size()
    assert D_padded.dim() == 3, D_padded.size()
    assert Q.size(0) in [1, D_padded.size(0)]

    scores = D_padded @ Q.to(dtype=D_padded.dtype).permute(0, 2, 1)

    return colbert_score_reduce(scores, D_mask, config)


_segmented_maxsim = None


def colbert_score_packed(Q, D_packed, D_lengths, config=ColBERTConfig()):
    """
    Works with a single query only.
    """

    use_gpu = config.total_visible_gpus > 0

    if use_gpu:
        Q, D_packed, D_lengths = Q.cuda(), D_packed.cuda(), D_lengths.cuda()

    Q = Q.squeeze(0)

    assert Q.dim() == 2, Q.size()
    assert D_packed.dim() == 2, D_packed.size()

    scores = D_packed @ Q.to(dtype=D_packed.dtype).T

    if use_gpu:
        scores_padded, scores_mask = StridedTensor(
            scores, D_lengths, use_gpu=use_gpu
        ).as_padded_tensor()

        return colbert_score_reduce(scores_padded, scores_mask, config)
    else:
        global _segmented_maxsim
        if _segmented_maxsim is None:
            _segmented_maxsim = try_load_torch_extensions(use_gpu)

        # _ = ColBERT(config)
        return _segmented_maxsim(scores, D_lengths)



from __future__ import annotations

import os
import pathlib
from math import ceil

import torch
from torch.utils.cpp_extension import load

from pylate.indexes.stanford_nlp.indexing.codecs.residual_embeddings_strided import (
    ResidualEmbeddingsStrided,
)
from pylate.indexes.stanford_nlp.modeling.colbert import (
    colbert_score,
    colbert_score_packed,
    colbert_score_reduce,
)
from pylate.indexes.stanford_nlp.search.candidate_generation import CandidateGeneration
from pylate.indexes.stanford_nlp.search.index_loader import IndexLoader
from pylate.indexes.stanford_nlp.search.strided_tensor import StridedTensor
from pylate.indexes.stanford_nlp.utils.utils import print_message


class IndexScorer(IndexLoader, CandidateGeneration):
    def __init__(self, index_path, use_gpu=True, load_index_with_mmap=False):
        super().__init__(
            index_path=index_path,
            use_gpu=use_gpu,
            load_index_with_mmap=load_index_with_mmap,
        )

        IndexScorer.try_load_torch_extensions(use_gpu)

        self.set_embeddings_strided()

    @classmethod
    def try_load_torch_extensions(cls, use_gpu):
        if hasattr(cls, "loaded_extensions") or use_gpu:
            return

        print_message(
            "Loading filter_pids_cpp extension (set COLBERT_LOAD_TORCH_EXTENSION_VERBOSE=True for more info)..."
        )
        filter_pids_cpp = load(
            name="filter_pids_cpp",
            sources=[
                os.path.join(
                    pathlib.Path(__file__).parent.resolve(), "filter_pids.cpp"
                ),
            ],
            extra_cflags=["-O3"],
            verbose=os.getenv("COLBERT_LOAD_TORCH_EXTENSION_VERBOSE", "False")
            == "True",
        )
        cls.filter_pids = filter_pids_cpp.filter_pids_cpp

        print_message(
            "Loading decompress_residuals_cpp extension (set COLBERT_LOAD_TORCH_EXTENSION_VERBOSE=True for more info)..."
        )
        decompress_residuals_cpp = load(
            name="decompress_residuals_cpp",
            sources=[
                os.path.join(
                    pathlib.Path(__file__).parent.resolve(), "decompress_residuals.cpp"
                ),
            ],
            extra_cflags=["-O3"],
            verbose=os.getenv("COLBERT_LOAD_TORCH_EXTENSION_VERBOSE", "False")
            == "True",
        )
        cls.decompress_residuals = decompress_residuals_cpp.decompress_residuals_cpp

        cls.loaded_extensions = True

    def set_embeddings_strided(self):
        if self.load_index_with_mmap:
            assert self.num_chunks == 1
            self.offsets = torch.cumsum(self.doclens, dim=0)
            self.offsets = torch.cat((torch.zeros(1, dtype=torch.int64), self.offsets))
        else:
            self.embeddings_strided = ResidualEmbeddingsStrided(
                self.codec, self.embeddings, self.doclens
            )
            self.offsets = self.embeddings_strided.codes_strided.offsets

    def lookup_pids(self, passage_ids, out_device="cuda", return_mask=False):
        return self.embeddings_strided.lookup_pids(passage_ids, out_device)

    def retrieve(self, config, Q):
        pids, centroid_scores = self.generate_candidates(config, Q)

        return pids, centroid_scores

    def embedding_ids_to_pids(self, embedding_ids):
        all_pids = torch.unique(self.emb2pid[embedding_ids.long()].cuda(), sorted=False)
        return all_pids

    def rank(self, config, Q, filter_fn=None, pids=None):
        with torch.inference_mode():
            if pids is None:
                pids, centroid_scores = self.retrieve(config, Q)
            else:
                pids = torch.tensor(pids, dtype=torch.int32, device=Q.device)
                centroid_scores = None

            if filter_fn is not None:
                filtered_pids = filter_fn(pids)
                assert isinstance(filtered_pids, torch.Tensor), type(filtered_pids)
                assert filtered_pids.dtype == pids.dtype, (
                    f"filtered_pids.dtype={filtered_pids.dtype}, pids.dtype={pids.dtype}"
                )
                assert filtered_pids.device == pids.device, (
                    f"filtered_pids.device={filtered_pids.device}, pids.device={pids.device}"
                )
                pids = filtered_pids
                if len(pids) == 0:
                    return [], []

            scores, pids = self.score_pids(config, Q, pids, centroid_scores)

            scores_sorter = scores.sort(descending=True)
            pids, scores = (
                pids[scores_sorter.indices].tolist(),
                scores_sorter.values.tolist(),
            )

            return pids, scores

    def score_pids(self, config, Q, pids, centroid_scores):
        """
        Always supply a flat list or tensor for `pids`.

        Supply sizes Q = (1 | num_docs, *, dim) and D = (num_docs, *, dim).
        If Q.size(0) is 1, the matrix will be compared with all passages.
        Otherwise, each query matrix will be compared against the *aligned* passage.
        """

        batch_size = config.search_batch_size

        if centroid_scores is not None:
            if self.use_gpu:
                centroid_scores = centroid_scores.cuda()

            idx = centroid_scores.max(-1).values >= config.centroid_score_threshold

            if self.use_gpu:
                approx_scores = []

                # Filter docs using pruned centroid scores
                for i in range(0, ceil(len(pids) / batch_size)):
                    pids_ = pids[i * batch_size : (i + 1) * batch_size]
                    codes_packed, codes_lengths = self.embeddings_strided.lookup_codes(
                        pids_
                    )
                    idx_ = idx[codes_packed.long()]
                    pruned_codes_strided = StridedTensor(
                        idx_, codes_lengths, use_gpu=self.use_gpu
                    )
                    pruned_codes_padded, pruned_codes_mask = (
                        pruned_codes_strided.as_padded_tensor()
                    )
                    pruned_codes_lengths = (
                        pruned_codes_padded * pruned_codes_mask
                    ).sum(dim=1)
                    codes_packed_ = codes_packed[idx_]
                    approx_scores_ = centroid_scores[codes_packed_.long()]
                    if approx_scores_.shape[0] == 0:
                        approx_scores.append(
                            torch.zeros(
                                (len(pids_),), dtype=approx_scores_.dtype
                            ).cuda()
                        )
                        continue
                    approx_scores_strided = StridedTensor(
                        approx_scores_, pruned_codes_lengths, use_gpu=self.use_gpu
                    )
                    approx_scores_padded, approx_scores_mask = (
                        approx_scores_strided.as_padded_tensor()
                    )
                    approx_scores_ = colbert_score_reduce(
                        approx_scores_padded, approx_scores_mask, config
                    )
                    approx_scores.append(approx_scores_)
                approx_scores = torch.cat(approx_scores, dim=0)
                assert approx_scores.is_cuda, approx_scores.device
                if config.ndocs < len(approx_scores):
                    pids = pids[torch.topk(approx_scores, k=config.ndocs).indices]

                # Filter docs using full centroid scores
                codes_packed, codes_lengths = self.embeddings_strided.lookup_codes(pids)
                approx_scores = centroid_scores[codes_packed.long()]
                approx_scores_strided = StridedTensor(
                    approx_scores, codes_lengths, use_gpu=self.use_gpu
                )
                approx_scores_padded, approx_scores_mask = (
                    approx_scores_strided.as_padded_tensor()
                )
                approx_scores = colbert_score_reduce(
                    approx_scores_padded, approx_scores_mask, config
                )
                if config.ndocs // 4 < len(approx_scores):
                    pids = pids[
                        torch.topk(approx_scores, k=(config.ndocs // 4)).indices
                    ]
            else:
                pids = IndexScorer.filter_pids(
                    pids,
                    centroid_scores,
                    self.embeddings.codes,
                    self.doclens,
                    self.offsets,
                    idx,
                    config.ndocs,
                )

        # Rank final list of docs using full approximate embeddings (including residuals)
        if self.use_gpu:
            D_packed, D_mask = self.lookup_pids(pids)
        else:
            D_packed = IndexScorer.decompress_residuals(
                pids,
                self.doclens,
                self.offsets,
                self.codec.bucket_weights,
                self.codec.reversed_bit_map,
                self.codec.decompression_lookup_table,
                self.embeddings.residuals,
                self.embeddings.codes,
                self.codec.centroids,
                self.codec.dim,
                self.codec.nbits,
            )
            D_packed = torch.nn.functional.normalize(
                D_packed.to(torch.float32), p=2, dim=-1
            )
            D_mask = self.doclens[pids.long()]

        if Q.size(0) == 1:
            return colbert_score_packed(Q, D_packed, D_mask, config), pids

        D_strided = StridedTensor(D_packed, D_mask, use_gpu=self.use_gpu)
        D_padded, D_lengths = D_strided.as_padded_tensor()

        return colbert_score(Q, D_padded, D_lengths, config), pids



"""Evaluation script for BEIR datasets with a PLAID index."""

from __future__ import annotations

import argparse

from pylate import evaluation, indexes, models, retrieve

if __name__ == "__main__":
    query_len = {
        "quora": 32,
        "climate-fever": 64,
        "nq": 32,
        "msmarco": 32,
        "hotpotqa": 32,
        "nfcorpus": 32,
        "scifact": 48,
        "trec-covid": 48,
        "fiqa": 32,
        "arguana": 64,
        "scidocs": 48,
        "dbpedia-entity": 32,
        "webis-touche2020": 32,
        "fever": 32,
        "cqadupstack/android": 32,
        "cqadupstack/english": 32,
        "cqadupstack/gaming": 32,
        "cqadupstack/gis": 32,
        "cqadupstack/mathematica": 32,
        "cqadupstack/physics": 32,
        "cqadupstack/programmers": 32,
        "cqadupstack/stats": 32,
        "cqadupstack/tex": 32,
        "cqadupstack/unix": 32,
        "cqadupstack/webmasters": 32,
        "cqadupstack/wordpress": 32,
    }

    # Parse dataset_name from command line arguments
    parser = argparse.ArgumentParser(description="Dataset name")
    parser.add_argument(
        "--dataset_name",
        type=str,
        default="nfcorpus",
        help="Name of the dataset to evaluate on (default: 'fiqa')",
    )
    args = parser.parse_args()
    dataset_name = args.dataset_name
    model_name = "lightonai/GTE-ModernColBERT-v1"
    model = models.ColBERT(
        model_name_or_path=model_name,
        document_length=300,
        query_length=query_len.get(dataset_name),
    )

    if "cqadupstack" in dataset_name:
        # Download dataset if not already downloaded
        from beir import util

        data_path = util.download_and_unzip(
            url="https://public.ukp.informatik.tu-darmstadt.de/thakur/BEIR/datasets/cqadupstack.zip",
            out_dir="./evaluation_datasets/",
        )
        documents, queries, qrels = evaluation.load_custom_dataset(
            f"evaluation_datasets/{dataset_name}",
            split="test",
        )
        dataset_name = dataset_name.replace("/", "_")
    else:
        documents, queries, qrels = evaluation.load_beir(
            dataset_name=dataset_name,
            split="dev" if "msmarco" in dataset_name else "test",
        )

    index = indexes.PLAID(
        override=True,
        index_name=f"{dataset_name}_{model_name.split('/')[-1]}",
    )

    retriever = retrieve.ColBERT(index=index)

    documents_embeddings = model.encode(
        sentences=[document["text"] for document in documents],
        batch_size=2000,
        is_query=False,
        show_progress_bar=True,
    )

    index.add_documents(
        documents_ids=[document["id"] for document in documents],
        documents_embeddings=documents_embeddings,
    )
    queries_embeddings = model.encode(
        sentences=list(queries.values()),
        is_query=True,
        show_progress_bar=True,
        batch_size=32,
    )

    scores = retriever.retrieve(queries_embeddings=queries_embeddings, k=20)

    # Remove query_id from scores, needed for FiQA dataset
    for (query_id, query), query_scores in zip(queries.items(), scores):
        for score in query_scores:
            if score["id"] == query_id:
                # Remove the query_id from the score
                query_scores.remove(score)

    evaluation_scores = evaluation.evaluate(
        scores=scores,
        qrels=qrels,
        queries=list(queries.keys()),
        # queries=queries,
        metrics=["map", "ndcg@10", "ndcg@100", "recall@10", "recall@100"],
    )

    print(evaluation_scores)



from __future__ import annotations

import torch


class IndexManager:
    def __init__(self, dim):
        self.dim = dim

    def save(self, tensor, path_prefix):
        torch.save(tensor, path_prefix)


def load_index_part(filename, verbose=True):
    part = torch.load(filename)

    if isinstance(part, list):  # for backward compatibility
        part = torch.cat(part)

    return part



from __future__ import annotations

from .scores import colbert_kd_scores, colbert_scores, colbert_scores_pairwise
from .similarity_functions import SimilarityFunction

__all__ = [
    "colbert_scores",
    "colbert_scores_pairwise",
    "colbert_kd_scores",
    "SimilarityFunction",
]



from __future__ import annotations

import os
import shutil

import pytest
from huggingface_hub.constants import HUGGINGFACE_HUB_CACHE

from pylate import models


@pytest.fixture(scope="function")
def cleanup_hf_cache():
    """A pytest fixture that runs after each test function.

    It completely removes the Hugging Face Hub cache directory.
    """

    yield

    print("Cleaning up Hugging Face cache...")
    if os.path.exists(HUGGINGFACE_HUB_CACHE):
        try:
            shutil.rmtree(HUGGINGFACE_HUB_CACHE)
            print("Hugging Face cache deleted successfully.")
        except OSError as e:
            print(f"Error deleting cache directory {HUGGINGFACE_HUB_CACHE}: {e}")
    else:
        print("Cache directory not found, skipping cleanup.")


@pytest.mark.model_loading
@pytest.mark.flaky(reruns=3, reruns_delay=5)
@pytest.mark.usefixtures("cleanup_hf_cache")
@pytest.mark.parametrize(
    "model_name_or_path, revision, query_prefix, document_prefix, max_seq_length, query_length, config",
    [
        (
            "colbert-ir/colbertv2.0",
            "c1e84128e85ef755c096a95bdb06b47793b13acf",
            "[Q]\\s",
            "[D]\\s",
            512,
            32,
            {},
        ),
        (
            "nanoColBERT/ColBERTv1",
            "547fc8b8a87b90a53d1d3f13548aeb5f4caf77c1",
            "[Q]\\s",
            "[D]\\s",
            512,
            32,
            {},
        ),
        (
            "samheym/GerColBERT",
            "d84587b4fd31d66039958d9552f429e7f1a288e7",
            "[unused0]",
            "[unused1]",
            512,
            32,
            {},
        ),
        (
            "lightonai/colbertv2.0",
            "29475aaab88a990378c0d04a63ed819d5c3ba830",
            "[unused0]",
            "[unused1]",
            512,
            32,
            {},
        ),
        (
            "jinaai/jina-colbert-v2",
            "4552c4dc1ffd7d7a635b6a41a1077fe9c9cdd974",
            "[QueryMarker]",
            "[DocumentMarker]",
            8194,
            32,
            {
                "query_prefix": "[QueryMarker]",
                "document_prefix": "[DocumentMarker]",
                "attend_to_expansion_tokens": True,
                "trust_remote_code": True,
            },
        ),
        (
            "antoinelouis/colbert-xm",
            "960de711799d210957d18df59c14c59a439b608a",
            "[unused0]",
            "[unused1]",
            514,
            32,
            {},
        ),
    ],
)
def test_load_model(
    model_name_or_path: str,
    revision: str,
    query_prefix: str,
    document_prefix: str,
    max_seq_length: int,
    query_length: int,
    config: dict,
):
    pylate_model = models.ColBERT(
        model_name_or_path=model_name_or_path, revision=revision, **config
    )

    assert hasattr(pylate_model, "encode")
    assert pylate_model.model_card_data.base_model_revision == revision
    assert pylate_model.model_card_data.library_name == "PyLate"
    assert pylate_model.similarity_fn_name == "MaxSim"
    assert pylate_model.query_prefix == (query_prefix.replace("\\s", " "))
    assert pylate_model.document_prefix == (document_prefix.replace("\\s", " "))
    assert pylate_model.max_seq_length == max_seq_length
    assert pylate_model.query_length == query_length



from __future__ import annotations

import logging
from typing import Sequence

import torch
import torch.distributed as dist

not_init_warning = True
logger = logging.getLogger(__name__)

_has_warned_dist_not_initialized = False


def all_gather(tensor: torch.Tensor) -> Sequence[torch.Tensor]:
    """Gathers a tensor from each distributed rank into a list. The tensor for the local rank is the original one, with the gradients while the others have no gradients.

    - If torch.distributed is available and initialized:
      1. Creates a list of tensors (each sized like the input `tensor`).
      2. Gathers tensors from each rank into that list.
      3. Replaces the local tensor in the list with the original tensor that retains gradients.

    - If torch.distributed is either unavailable, uninitialized, or
      `world_size == 1`, it returns a list containing only the
      original tensor and throws a warning to notify the user (helpful when using a single GPU setup).

    Parameters
    ----------
    tensor:
        The input tensor to be gathered from each rank.

    Returns
    -------
    Sequence:
        A list of tensors collected from each rank. On a single GPU or when distributed is uninitialized, the list will contain only the original tensor.

    """
    global _has_warned_dist_not_initialized

    # Check if torch.distributed is properly available and initialized.
    if dist.is_available() and dist.is_initialized():
        world_size = dist.get_world_size()
        gathered_tensors = [torch.zeros_like(tensor) for _ in range(world_size)]

        # Perform all_gather.
        dist.all_gather(gathered_tensors, tensor)

        # Replace local rank's tensor with the original (retaining gradients).
        local_rank = dist.get_rank()
        gathered_tensors[local_rank] = tensor
        return gathered_tensors

    # Warn once about uninitialized or single-GPU usage.
    if not _has_warned_dist_not_initialized:
        warning = """
            Trying to gather while torch.distributed is not available or has not been initialized,
             returning the original (local) tensor. This is expected if you are
             only using one GPU; consider not using gathering to remove this warning.
       """
        logger.warning(warning)
        _has_warned_dist_not_initialized = True

    return [tensor]


def all_gather_with_gradients(tensor: torch.Tensor) -> Sequence[torch.Tensor]:
    """Gathers a tensor from each distributed rank into a list. All the tensors will retain gradients.
    This is the same as `all_gather`, but all the tensors will retain gradients and is used to compute contrastive with local queries only to lower the memory usage, see https://github.com/mlfoundations/open_clip/issues/616

    - If torch.distributed is available and initialized, gather all the tensors (with gradients) from each rank into a list

    - If torch.distributed is either unavailable, uninitialized, or
      `world_size == 1`, it returns a list containing only the
      original tensor and throws a warning to notify the user (helpful when using a single GPU setup).

    Parameters
    ----------
    tensor:
        The input tensor to be gathered from each rank.

    Returns
    -------
    Sequence:
        A list of tensors collected from each rank. On a single GPU or when distributed is uninitialized, the list will contain only the original tensor.

    """
    global _has_warned_dist_not_initialized

    # Check if torch.distributed is properly available and initialized.
    if dist.is_available() and dist.is_initialized():
        tensor = dist.nn.all_gather(tensor)
        return tensor

    # Warn once about uninitialized or single-GPU usage.
    if not _has_warned_dist_not_initialized:
        warning = """
            Trying to gather while torch.distributed is not available or has not been initialized,
             returning the original (local) tensor. This is expected if you are
             only using one GPU; consider not using gathering to remove this warning.
       """
        logger.warning(warning)
        _has_warned_dist_not_initialized = True

    return [tensor]


def get_rank() -> int:
    """Returns the current rank in a distributed training."""
    # Check if torch.distributed is properly available and initialized.
    if dist.is_available() and dist.is_initialized():
        return dist.get_rank()
    return 0


def get_world_size() -> int:
    """Returns the world size in a distributed training."""
    # Check if torch.distributed is properly available and initialized.
    if dist.is_available() and dist.is_initialized():
        return dist.get_world_size()
    return 1



from __future__ import annotations

import os
from dataclasses import dataclass

import torch

import __main__
from pylate.indexes.stanford_nlp.utils.utils import timestamp

from .core_config import DefaultVal


@dataclass
class RunSettings:
    """
    The defaults here have a special status in Run(), which initially calls assign_defaults(),
    so these aren't soft defaults in that specific context.
    """

    overwrite: bool = DefaultVal(False)

    root: str = DefaultVal(os.path.join(os.getcwd(), "experiments"))
    experiment: str = DefaultVal("default")

    index_root: str = DefaultVal(None)
    name: str = DefaultVal(timestamp(daydir=True))

    rank: int = DefaultVal(0)
    nranks: int = DefaultVal(1)
    amp: bool = DefaultVal(True)

    total_visible_gpus = torch.cuda.device_count()
    gpus: int = DefaultVal(total_visible_gpus)

    avoid_fork_if_possible: bool = DefaultVal(False)

    @property
    def gpus_(self):
        value = self.gpus

        if isinstance(value, int):
            value = list(range(value))

        if isinstance(value, str):
            value = value.split(",")

        value = list(map(int, value))
        value = sorted(list(set(value)))

        assert all(
            device_idx in range(0, self.total_visible_gpus) for device_idx in value
        ), value

        return value

    @property
    def index_root_(self):
        return self.index_root or os.path.join(self.root, self.experiment, "indexes/")

    @property
    def script_name_(self):
        if "__file__" in dir(__main__):
            cwd = os.path.abspath(os.getcwd())
            script_path = os.path.abspath(__main__.__file__)
            root_path = os.path.abspath(self.root)

            if script_path.startswith(cwd):
                script_path = script_path[len(cwd) :]

            else:
                try:
                    commonpath = os.path.commonpath([script_path, root_path])
                    script_path = script_path[len(commonpath) :]
                except Exception:
                    pass

            assert script_path.endswith(".py")
            script_name = script_path.replace("/", ".").strip(".")[:-3]

            assert len(script_name) > 0, (script_name, script_path, cwd)

            return script_name

        return "none"

    @property
    def path_(self):
        return os.path.join(self.root, self.experiment, self.script_name_, self.name)

    @property
    def device_(self):
        return self.gpus_[self.rank % self.nranks]


@dataclass
class ResourceSettings:
    collection: str = DefaultVal(None)
    index_name: str = DefaultVal(None)


@dataclass
class DocSettings:
    dim: int = DefaultVal(128)


@dataclass
class IndexingSettings:
    index_path: str = DefaultVal(None)

    index_bsize: int = DefaultVal(64)

    nbits: int = DefaultVal(1)

    kmeans_niters: int = DefaultVal(4)

    resume: bool = DefaultVal(False)

    use_triton: bool | None = DefaultVal(None)

    @property
    def index_path_(self):
        return self.index_path or os.path.join(self.index_root_, self.index_name)


@dataclass
class SearchSettings:
    ncells: int = DefaultVal(None)
    centroid_score_threshold: float = DefaultVal(None)
    ndocs: int = DefaultVal(None)
    load_index_with_mmap: bool = DefaultVal(False)
    search_batch_size: int = DefaultVal(2**18)



from __future__ import annotations

import torch
from datasets import load_dataset
from sentence_transformers import (
    SentenceTransformerTrainer,
    SentenceTransformerTrainingArguments,
)

from pylate import evaluation, losses, models, utils

# Define model parameters for contrastive training
model_name = "bert-base-uncased"  # Choose the pre-trained model you want to use as base
batch_size = 32  # Larger batch size often improves results, but requires more memory

num_train_epochs = 1  # Adjust based on your requirements
# Set the run name for logging and output directory
run_name = "contrastive-bert-base-uncased"
output_dir = f"output/{run_name}"

# 1. Here we define our ColBERT model. If not a ColBERT model, will add a linear layer to the base encoder.
model = models.ColBERT(model_name_or_path=model_name)

# Compiling the model makes the training faster
model = torch.compile(model)

# Load dataset
dataset = load_dataset("sentence-transformers/msmarco-bm25", "triplet", split="train")
# Split the dataset (this dataset does not have a validation set, so we split the training set)
splits = dataset.train_test_split(test_size=0.01)
train_dataset = splits["train"]
eval_dataset = splits["test"]

# Define the loss function
train_loss = losses.Contrastive(model=model)

# Initialize the evaluator
dev_evaluator = evaluation.ColBERTTripletEvaluator(
    anchors=eval_dataset["query"],
    positives=eval_dataset["positive"],
    negatives=eval_dataset["negative"],
)

# Configure the training arguments (e.g., batch size, evaluation strategy, logging steps)
args = SentenceTransformerTrainingArguments(
    output_dir=output_dir,
    num_train_epochs=num_train_epochs,
    per_device_train_batch_size=batch_size,
    per_device_eval_batch_size=batch_size,
    fp16=True,  # Set to False if you get an error that your GPU can't run on FP16
    bf16=False,  # Set to True if you have a GPU that supports BF16
    run_name=run_name,  # Will be used in W&B if `wandb` is installed
    learning_rate=3e-6,
)

# Initialize the trainer for the contrastive training
trainer = SentenceTransformerTrainer(
    model=model,
    args=args,
    train_dataset=train_dataset,
    eval_dataset=eval_dataset,
    loss=train_loss,
    evaluator=dev_evaluator,
    data_collator=utils.ColBERTCollator(model.tokenize),
)
# Start the training process
trainer.train()



from __future__ import annotations

import logging
import os

import numpy as np
import torch
from fast_plaid import search
from sqlitedict import SqliteDict

from ..rank import RerankResult
from .base import Base

logger = logging.getLogger(__name__)


def convert_embeddings_to_torch(
    embeddings: np.ndarray | torch.Tensor | list,
) -> list[torch.Tensor]:
    """Convert embeddings to list of torch tensors as expected by fast-plaid."""
    if isinstance(embeddings, list):
        if len(embeddings) == 0:
            return []
        if isinstance(embeddings[0], torch.Tensor):
            return embeddings
        elif isinstance(embeddings[0], np.ndarray):
            return [torch.from_numpy(emb) for emb in embeddings]

    if isinstance(embeddings, np.ndarray):
        if len(embeddings.shape) == 3:  # batch_size, n_tokens, embedding_size
            return [torch.from_numpy(embeddings[i]) for i in range(embeddings.shape[0])]
        elif len(embeddings.shape) == 2:  # n_tokens, embedding_size
            return [torch.from_numpy(embeddings)]

    if isinstance(embeddings, torch.Tensor):
        if len(embeddings.shape) == 3:  # batch_size, n_tokens, embedding_size
            return [embeddings[i] for i in range(embeddings.shape[0])]
        elif len(embeddings.shape) == 2:  # n_tokens, embedding_size
            return [embeddings]

    return embeddings


class FastPlaid(Base):
    """FastPlaid index using the fast-plaid backend for high-performance multi-vector search.

    Parameters
    ----------
    index_folder
        The folder where the index will be stored.
    index_name
        The name of the index.
    override
        Whether to override the collection if it already exists.
    nbits
        The number of bits to use for product quantization.
        Lower values mean more compression and potentially faster searches but can reduce accuracy.
    kmeans_niters
        The number of iterations for the K-means algorithm used during index creation.
        This influences the quality of the initial centroid assignments.
    max_points_per_centroid
        The maximum number of points (token embeddings) that can be assigned to a single centroid during K-means.
        This helps in balancing the clusters.
    n_ivf_probe
        The number of inverted file list "probes" to perform during the search.
        This parameter controls the number of clusters to search within the index for each query.
        Higher values improve recall but increase search time.
    n_full_scores
        The number of candidate documents for which full (re-ranked) scores are computed.
        This is a crucial parameter for accuracy; higher values lead to more accurate results but increase computation.
    n_samples_kmeans
        The number of samples to use for K-means clustering.
        If None, it defaults to a value based on the number of documents.
        This parameter can be adjusted to balance between speed, memory usage and clustering quality.
    batch_size
        The internal batch size used for processing queries.
        A larger batch size might improve throughput on powerful GPUs but can consume more memory.
    show_progress
        If set to True, a progress bar will be displayed during search operations.
    device
        Specifies the device(s) to use for computation.
        If None (default) and CUDA is available, it defaults to "cuda".
        If CUDA is not available, it defaults to "cpu".
        Can be a single device string (e.g., "cuda:0" or "cpu").
        Can be a list of device strings (e.g., ["cuda:0", "cuda:1"]).
    use_triton
        Whether to use triton kernels when computing kmeans using fast-plaid. Triton kernels are faster, but yields some variance due to race condition, set to false to get 100% reproducable results. If unset, will use triton kernels if possible.

    """

    def __init__(
        self,
        index_folder: str = "indexes",
        index_name: str = "fast_plaid",
        override: bool = False,
        nbits: int = 4,
        kmeans_niters: int = 4,
        max_points_per_centroid: int = 256,
        n_ivf_probe: int = 8,
        n_full_scores: int = 8192,
        n_samples_kmeans: int | None = None,
        batch_size: int = 1 << 18,
        show_progress: bool = True,
        device: str | list[str] | None = None,
        use_triton: bool | None = None,
    ) -> None:
        self.index_folder = index_folder
        self.index_name = index_name
        self.nbits = nbits
        self.kmeans_niters = kmeans_niters
        self.max_points_per_centroid = max_points_per_centroid
        self.n_ivf_probe = n_ivf_probe
        self.n_full_scores = n_full_scores
        self.n_samples_kmeans = n_samples_kmeans
        self.batch_size = batch_size
        self.show_progress = show_progress
        self.device = device
        self.use_triton = use_triton

        # Create the index directory structure
        self.index_path = os.path.join(index_folder, index_name)
        self.fast_plaid_index_path = os.path.join(self.index_path, "fast_plaid_index")

        if not os.path.exists(index_folder):
            os.makedirs(index_folder)
        if not os.path.exists(self.index_path):
            os.makedirs(self.index_path)

        # SQLite mappings for document IDs
        self.documents_ids_to_plaid_ids_path = os.path.join(
            self.index_path, "documents_ids_to_plaid_ids.sqlite"
        )
        self.plaid_ids_to_documents_ids_path = os.path.join(
            self.index_path, "plaid_ids_to_documents_ids.sqlite"
        )

        # Initialize or load the fast-plaid index
        self.fast_plaid = search.FastPlaid(
            index=self.fast_plaid_index_path, device=device
        )

        if override:
            # Remove existing SQLite mappings
            if os.path.exists(self.documents_ids_to_plaid_ids_path):
                os.remove(self.documents_ids_to_plaid_ids_path)
            if os.path.exists(self.plaid_ids_to_documents_ids_path):
                os.remove(self.plaid_ids_to_documents_ids_path)
            self.is_indexed = False
        else:
            # Check if index already exists
            documents_ids_to_plaid_ids = self._load_documents_ids_to_plaid_ids()
            self.is_indexed = len(documents_ids_to_plaid_ids) > 0
            documents_ids_to_plaid_ids.close()

    def _load_documents_ids_to_plaid_ids(self) -> SqliteDict:
        """Load the SQLite database that maps document IDs to PLAID IDs."""
        return SqliteDict(self.documents_ids_to_plaid_ids_path, outer_stack=False)

    def _load_plaid_ids_to_documents_ids(self) -> SqliteDict:
        """Load the SQLite database that maps PLAID IDs to document IDs."""
        return SqliteDict(self.plaid_ids_to_documents_ids_path, outer_stack=False)

    def add_documents(
        self,
        documents_ids: str | list[str],
        documents_embeddings: list[np.ndarray | torch.Tensor],
        **kwargs,
    ) -> "FastPlaid":
        """Add documents to the index using fast-plaid backend."""
        if isinstance(documents_ids, str):
            documents_ids = [documents_ids]

        # Convert embeddings to torch tensors
        documents_embeddings_torch = convert_embeddings_to_torch(documents_embeddings)

        # Load SQLite mappings
        documents_ids_to_plaid_ids = self._load_documents_ids_to_plaid_ids()
        plaid_ids_to_documents_ids = self._load_plaid_ids_to_documents_ids()

        # Get the current number of documents for ID assignment
        current_max_id = (
            max([int(k) for k in plaid_ids_to_documents_ids.keys()])
            if plaid_ids_to_documents_ids
            else -1
        )

        if not self.is_indexed:
            # Create new index
            self.fast_plaid.create(
                documents_embeddings=documents_embeddings_torch,
                kmeans_niters=self.kmeans_niters,
                max_points_per_centroid=self.max_points_per_centroid,
                nbits=self.nbits,
                n_samples_kmeans=self.n_samples_kmeans,
                use_triton_kmeans=self.use_triton,
            )
            plaid_ids = list(range(len(documents_embeddings_torch)))
            self.is_indexed = True
        else:
            # Update existing index
            logger.warning(
                "Adding documents to existing index. This uses fast-plaid's update method "
                "which does not recompute centroids and may result in slightly lower accuracy."
            )
            self.fast_plaid.update(documents_embeddings=documents_embeddings_torch)
            # Assign new plaid IDs starting from current_max_id + 1
            plaid_ids = list(
                range(
                    current_max_id + 1,
                    current_max_id + 1 + len(documents_embeddings_torch),
                )
            )

        # Store mappings
        for plaid_id, document_id in zip(plaid_ids, documents_ids):
            documents_ids_to_plaid_ids[document_id] = plaid_id
            plaid_ids_to_documents_ids[plaid_id] = document_id

        documents_ids_to_plaid_ids.commit()
        documents_ids_to_plaid_ids.close()

        plaid_ids_to_documents_ids.commit()
        plaid_ids_to_documents_ids.close()

        return self

    def remove_documents(self, documents_ids: list[str]) -> "FastPlaid":
        """Remove documents from the index.

        Note: Fast-plaid does not support direct document removal.
        This method removes the document mappings but the embeddings remain in the index.
        For complete removal, consider rebuilding the index without the unwanted documents.

        Parameters
        ----------
        documents_ids
            The document IDs to remove.
        """
        logger.warning(
            "Fast-plaid does not support direct document removal from the index. "
            "This will only remove the document ID mappings. For complete removal, "
            "consider rebuilding the index without the unwanted documents."
        )

        documents_ids_to_plaid_ids = self._load_documents_ids_to_plaid_ids()
        plaid_ids_to_documents_ids = self._load_plaid_ids_to_documents_ids()

        # Remove mappings
        for document_id in documents_ids:
            if document_id in documents_ids_to_plaid_ids:
                plaid_id = documents_ids_to_plaid_ids[document_id]
                del documents_ids_to_plaid_ids[document_id]
                del plaid_ids_to_documents_ids[plaid_id]

        documents_ids_to_plaid_ids.commit()
        documents_ids_to_plaid_ids.close()

        plaid_ids_to_documents_ids.commit()
        plaid_ids_to_documents_ids.close()

        return self

    def __call__(
        self,
        queries_embeddings: np.ndarray
        | torch.Tensor
        | list[np.ndarray]
        | list[torch.Tensor],
        k: int = 10,
        subset: list[list[str]] | list[str] | None = None,
    ) -> list[list[RerankResult]]:
        """Query the index for the nearest neighbors of the queries embeddings.

        Parameters
        ----------
        queries_embeddings
            The query embeddings. Can be numpy array, torch tensor, or list of numpy arrays/torch tensors.
        k
            The number of nearest neighbors to return.
        subset
            Optional subset of document IDs to restrict search to.
            Can be a single list (same filter for all queries) or
            list of lists (different filter per query).
            Document IDs should match the IDs used when adding documents.

        Returns
        -------
        List of lists containing dictionaries with 'id' and 'score' keys.
        """
        if not self.is_indexed:
            error = """
            The index is empty. Please add documents before querying.
            """
            raise ValueError(error)

        plaid_ids_to_documents_ids = self._load_plaid_ids_to_documents_ids()
        documents_ids_to_plaid_ids = self._load_documents_ids_to_plaid_ids()

        # Convert queries to torch tensor format expected by fast-plaid
        queries_embeddings = convert_embeddings_to_torch(queries_embeddings)

        # Convert subset from document IDs to plaid IDs if provided
        plaid_subset = None
        if subset is not None:
            if len(subset) == 0:
                # Empty list - return empty results
                plaid_subset = []
            elif isinstance(subset[0], list):
                # List of lists - different subset for each query
                plaid_subset = []
                for query_subset in subset:
                    query_plaid_ids = [
                        documents_ids_to_plaid_ids[doc_id]
                        for doc_id in query_subset
                        if doc_id in documents_ids_to_plaid_ids
                    ]
                    plaid_subset.append(query_plaid_ids)
            else:
                # Single list - same subset for all queries
                plaid_subset = [
                    documents_ids_to_plaid_ids[doc_id]
                    for doc_id in subset
                    if doc_id in documents_ids_to_plaid_ids
                ]

        # Perform search using fast-plaid
        search_results = self.fast_plaid.search(
            queries_embeddings=queries_embeddings,
            top_k=k,
            batch_size=self.batch_size,
            n_ivf_probe=self.n_ivf_probe,
            n_full_scores=self.n_full_scores,
            show_progress=self.show_progress,
            subset=plaid_subset,
        )

        # Convert results to expected format
        results = []
        for query_results in search_results:
            query_docs = []
            for plaid_id, score in query_results:
                if plaid_id in plaid_ids_to_documents_ids:
                    doc_id = plaid_ids_to_documents_ids[plaid_id]
                    query_docs.append(RerankResult(id=doc_id, score=float(score)))
            results.append(query_docs)

        plaid_ids_to_documents_ids.close()
        documents_ids_to_plaid_ids.close()
        return results

    def get_documents_embeddings(
        self, document_ids: list[list[str]]
    ) -> list[list[list[int | float]]]:
        """Get document embeddings by their IDs.

        Note: Fast-plaid does not provide direct access to document embeddings.
        This method is not implemented as the embeddings are stored in compressed form.
        """
        raise NotImplementedError(
            "Fast-plaid does not provide direct access to document embeddings. "
            "The embeddings are stored in compressed/quantized form and cannot be retrieved."
        )



from __future__ import annotations

from .cached_contrastive import CachedContrastive
from .contrastive import Contrastive
from .distillation import Distillation

__all__ = ["Contrastive", "Distillation", "CachedContrastive"]



from __future__ import annotations

from .collator import ColBERTCollator
from .distributed import all_gather, all_gather_with_gradients, get_rank, get_world_size
from .huggingface_models import HUGGINGFACE_MODELS
from .iter_batch import iter_batch
from .multi_process import _start_multi_process_pool
from .processing import KDProcessing
from .tensor import convert_to_tensor

__all__ = [
    "HUGGINGFACE_MODELS",
    "iter_batch",
    "convert_to_tensor",
    "ColBERTCollator",
    "KDProcessing",
    "_start_multi_process_pool",
    "all_gather",
    "all_gather_with_gradients",
    "get_rank",
    "get_world_size",
]



"""Evaluation script for the custom dataset."""

from __future__ import annotations

from pylate import evaluation, indexes, models, retrieve

model = models.ColBERT(
    model_name_or_path="lightonai/GTE-ModernColBERT-v1",
    document_length=300,
)
index = indexes.PLAID(override=True)
retriever = retrieve.ColBERT(index=index)

documents, queries, qrels = evaluation.load_custom_dataset(
    "datasets/miracl_fr", split="dev"
)

documents_embeddings = model.encode(
    sentences=[document["text"] for document in documents],
    batch_size=32,
    is_query=False,
    show_progress_bar=True,
)

index.add_documents(
    documents_ids=[document["id"] for document in documents],
    documents_embeddings=documents_embeddings,
)

queries_embeddings = model.encode(
    sentences=list(queries.values()),
    batch_size=32,
    is_query=True,
    show_progress_bar=True,
)

scores = retriever.retrieve(queries_embeddings=queries_embeddings, k=100)

evaluation_scores = evaluation.evaluate(
    scores=scores,
    qrels=qrels,
    queries=list(queries.keys()),
    metrics=["map", "ndcg@10", "ndcg@100", "recall@10", "recall@100"],
)

print(evaluation_scores)



from __future__ import annotations

from typing import Iterable

import torch
import torch.nn.functional as F
from torch import Tensor, nn

from ..models import ColBERT
from ..scores import colbert_scores
from ..utils import all_gather, all_gather_with_gradients, get_rank, get_world_size


def extract_skiplist_mask(
    sentence_features: Iterable[dict[str, torch.Tensor]],
    skiplist: list[int],
) -> list[torch.Tensor]:
    """Extracts the attention masks from the sentence features. We apply a skiplist mask to the documents.
    We skip the first sentence feature because it is the query.

    Examples
    --------
    >>> import torch

    >>> sentence_features = [
    ...     {
    ...         "input_ids": torch.tensor([[1, 2, 3, 4]]),
    ...         "attention_mask": torch.tensor([[1, 1, 1, 1]]),
    ...     },
    ...     {
    ...         "input_ids": torch.tensor([[1, 2, 3, 4]]),
    ...         "attention_mask": torch.tensor([[1, 1, 1, 1]]),
    ...     },
    ...     {
    ...         "input_ids": torch.tensor([[1, 2, 3, 4]]),
    ...         "attention_mask": torch.tensor([[1, 1, 1, 1]]),
    ...     },
    ... ]

    >>> extract_skiplist_mask(
    ...     sentence_features=sentence_features,
    ...     skiplist=[1, 2, 3],
    ... )
    [tensor([[True, True, True, True]]), tensor([[False, False, False,  True]]), tensor([[False, False, False,  True]])]

    """
    attention_masks = [
        sentence_feature["attention_mask"] for sentence_feature in sentence_features
    ]

    skiplist_masks = [
        torch.ones_like(sentence_features[0]["input_ids"], dtype=torch.bool)
    ]

    # We skip the first sentence feature because it is the query.
    skiplist_masks.extend(
        [
            ColBERT.skiplist_mask(
                input_ids=sentence_feature["input_ids"], skiplist=skiplist
            )
            for sentence_feature in sentence_features[1:]
        ]
    )

    return [
        torch.logical_and(skiplist_mask, attention_mask)
        for skiplist_mask, attention_mask in zip(skiplist_masks, attention_masks)
    ]


class Contrastive(nn.Module):
    """
    Contrastive loss. Expects as input two texts and a label of either 0 or 1. If the label == 1, then the distance between the
    two embeddings is reduced. If the label == 0, then the distance between the embeddings is increased.

    Parameters
    ----------
    model
        ColBERT model.
    score_metric
        ColBERT scoring function. Defaults to colbert_scores.
    size_average
        Average by the size of the mini-batch.
    gather_across_devices
        Whether to gather the embeddings across devices to have more in batch negatives. We recommand making sure the sampling across GPUs use the same dataset in case of multi-dataset training to make sure the negatives are plausible.

    Examples
    --------
    >>> from pylate import models, losses

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2", device="cpu"
    ... )

    >>> loss = losses.Contrastive(model=model)

    >>> anchor = model.tokenize([
    ...     "fruits are healthy.",
    ... ], is_query=True)

    >>> positive = model.tokenize([
    ...     "fruits are good for health.",
    ... ], is_query=False)

    >>> negative = model.tokenize([
    ...     "fruits are bad for health.",
    ... ], is_query=False)

    >>> sentence_features = [anchor, positive, negative]

    >>> loss = loss(sentence_features=sentence_features)
    >>> assert isinstance(loss.item(), float)

    """

    def __init__(
        self,
        model: ColBERT,
        score_metric=colbert_scores,
        size_average: bool = True,
        gather_across_devices: bool = False,
        temperature: float = 1.0,
    ) -> None:
        super(Contrastive, self).__init__()
        self.score_metric = score_metric
        self.model = model
        self.size_average = size_average
        self.gather_across_devices = gather_across_devices
        self.temperature = temperature

    def forward(
        self,
        sentence_features: Iterable[dict[str, Tensor]],
        labels: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Compute the Constrastive loss.

        Parameters
        ----------
        sentence_features
            List of tokenized sentences. The first sentence is the anchor and the rest are the positive and negative examples.
        labels
            The labels for the contrastive loss. Not used in this implementation, but kept for compatibility with Trainer.

        """
        embeddings = [
            torch.nn.functional.normalize(
                self.model(sentence_feature)["token_embeddings"], p=2, dim=-1
            )
            for sentence_feature in sentence_features
        ]
        # handle the model being wrapped in (D)DP and so require to access module first
        skiplist = (
            self.model.skiplist
            if hasattr(self.model, "skiplist")
            else self.model.module.skiplist
        )
        do_query_expansion = (
            self.model.do_query_expansion
            if hasattr(self.model, "do_query_expansion")
            else self.model.module.do_query_expansion
        )
        masks = extract_skiplist_mask(
            sentence_features=sentence_features, skiplist=skiplist
        )
        batch_size = embeddings[0].size(0)
        # create corresponding labels
        labels = torch.arange(0, batch_size, device=embeddings[0].device)
        # Possibly gather the embeddings across devices to have more in-batch negatives.
        if self.gather_across_devices:
            # Note that we only gather the documents embeddings and not the queries embeddings (embeddings[0]), but are keeping gradients. This is to lower the memory usage, see https://github.com/mlfoundations/open_clip/issues/616
            embeddings = [
                embeddings[0],
                *[
                    torch.cat(all_gather_with_gradients(embedding))
                    for embedding in embeddings[1:]
                ],
            ]
            # Masks [0] is the anchor mask so we do not need to gather it (even though we are not using it for now anyways)
            # Also, we do gather without gradients for the masks as we do not backpropagate through them
            masks = [
                masks[0],
                *[torch.cat(all_gather(mask)) for mask in masks[1:]],
            ]
            rank = get_rank()
            # Adjust the labels to match the gathered embeddings positions
            labels = labels + rank * batch_size
        # Note: the queries mask is not used, if added, take care that the expansion tokens are not masked from scoring (because they might be masked during encoding).
        # We might not need to compute the mask for queries but I let the logic there for now
        scores = torch.cat(
            [
                self.score_metric(
                    embeddings[0],
                    group_embeddings,
                    queries_mask=masks[0] if not do_query_expansion else None,
                    documents_mask=documents_masks,
                )
                for group_embeddings, documents_masks in zip(embeddings[1:], masks[1:])
            ],
            dim=1,
        )

        # compute constrastive loss using cross-entropy over the scores
        loss = F.cross_entropy(
            input=scores / self.temperature,
            target=labels,
            reduction="mean" if self.size_average else "sum",
        )

        # Scale by world size when gathering across device
        if self.gather_across_devices:
            loss *= get_world_size()
        return loss



from __future__ import annotations

import argparse
from typing import List

import batched
import uvicorn
from fastapi import FastAPI, HTTPException
from pydantic import BaseModel

from pylate import models

app = FastAPI()


class EmbeddingRequest(BaseModel):
    """PyDantic model for the requests sent to the server.

    Parameters
    ----------
    input
        The input(s) to encode.
    is_query
        A boolean indicating if the input is a query or a document.
    model
        The name model to use for encoding.
    """

    input: List[str] | str
    is_query: bool = True
    model: str = "lightonai/GTE-ModernColBERT-v1"


class EmbeddingResponse(BaseModel):
    """PyDantic model for the server answer to a call.

    Parameters
    ----------
    data
        A list of dictionaries containing the embeddings ("embedding" key) and the type of the object ("object" key, is always embedding).
    model
        The name of the model used for encoding.
    usage
        An approximation of the number of tokens used to generate the embeddings (computed by splitting the input sequences on spaces).
    """

    data: List[dict]
    model: str
    usage: dict


def wrap_encode_function(model, **kwargs):
    def wrapped_encode(sentences):
        return model.encode(sentences, **kwargs)

    return wrapped_encode


def parse_args():
    parser = argparse.ArgumentParser(
        description="Run FastAPI ColBERT serving server with specified host, port, and model."
    )
    parser.add_argument(
        "--host", type=str, default="0.0.0.0", help="Host to run the server on"
    )
    parser.add_argument(
        "--port", type=int, default=8002, help="Port to run the server on"
    )
    parser.add_argument(
        "--model",
        type=str,
        default="lightonai/GTE-ModernColBERT-v1",
        help="Model to serve, can be an HF model or a path to a model",
    )
    return parser.parse_args()


args = parse_args()

# We need to load the model here so it is shared for every request
model = models.ColBERT(args.model)
# We cannot create the function on the fly as the batching require to use the same function (memory address)
model.encode_query = batched.aio.dynamically(wrap_encode_function(model, is_query=True))
model.encode_document = batched.aio.dynamically(
    wrap_encode_function(model, is_query=False)
)


@app.post("/v1/embeddings", response_model=EmbeddingResponse)
async def create_embedding(request: EmbeddingRequest):
    """API endpoint that encode the elements of an EmbeddingRequest and returns an EmbeddingResponse.

    Parameters
    ----------
    request
        The EmbeddingRequest containing the elements to encode, the model to use, and whether the input is a query or a document.
    """
    if request.model != args.model:
        raise HTTPException(
            status_code=400,
            detail=f"Model not supported, the loaded model is {args.model}, but the request is for {request.model}",
        )
    try:
        if request.is_query:
            embeddings = await model.encode_query(
                request.input,
            )
        else:
            embeddings = await model.encode_document(
                request.input,
            )

        # Format response
        data = [
            {"object": "embedding", "embedding": embedding.tolist(), "index": i}
            for i, embedding in enumerate(embeddings)
        ]

        # Calculate token usage (approximate)
        total_tokens = sum(len(text.split()) for text in request.input)

        return EmbeddingResponse(
            data=data,
            model=request.model,
            usage={"prompt_tokens": total_tokens, "total_tokens": total_tokens},
        )
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    uvicorn.run(app, host=args.host, port=args.port)



from __future__ import annotations

import json
import logging
import os

import torch
from safetensors import safe_open
from safetensors.torch import load_model as load_safetensors_model
from sentence_transformers.models import Dense as DenseSentenceTransformer
from sentence_transformers.util import import_from_string
from torch import nn
from transformers.utils import cached_file

__all__ = ["Dense"]

logger = logging.getLogger(__name__)


class Dense(DenseSentenceTransformer):
    """Performs linear projection on the token embeddings to a lower dimension.

    Parameters
    ----------
    in_features
        Size of the embeddings in output of the transformer.
    out_features
        Size of the output embeddings after linear projection
    bias
        Add a bias vector
    init_weight
        Initial value for the matrix of the linear layer
    init_bias
        Initial value for the bias of the linear layer.
    use_residual
        Whether to use residual for the linear layer.

    Examples
    --------
    >>> from pylate import models

    >>> model = models.Dense(
    ...     in_features=768,
    ...     out_features=128,
    ... )

    >>> features = {
    ...     "token_embeddings": torch.randn(2, 768),
    ... }

    >>> projected_features = model(features)

    >>> assert projected_features["token_embeddings"].shape == (2, 128)
    >>> assert isinstance(model, DenseSentenceTransformer)

    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        activation_function=nn.Identity(),
        init_weight: torch.Tensor = None,
        init_bias: torch.Tensor = None,
        use_residual: bool = False,
    ) -> None:
        super(Dense, self).__init__(
            in_features, out_features, bias, activation_function, init_weight, init_bias
        )
        self.use_residual = use_residual
        if use_residual and self.in_features != self.out_features:
            self.residual = nn.Linear(self.in_features, self.out_features, bias=False)

    def forward(self, features: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Performs linear projection on the token embeddings."""
        token_embeddings = features["token_embeddings"]
        projected_embeddings = self.activation_function(self.linear(token_embeddings))
        if self.use_residual:
            residual_embeddings = token_embeddings
            if self.in_features != self.out_features:
                residual_embeddings = self.residual(token_embeddings)
            projected_embeddings = projected_embeddings + residual_embeddings
        features["token_embeddings"] = projected_embeddings
        return features

    @staticmethod
    def from_sentence_transformers(dense: DenseSentenceTransformer) -> "Dense":
        """Converts a SentenceTransformer Dense model to a Dense model.
        Our Dense model does not have the activation function.
        """
        config = dense.get_config_dict()
        config["activation_function"] = import_from_string(
            config["activation_function"]
        )()
        model = Dense(**config)
        model.load_state_dict(dense.state_dict())
        return model

    @staticmethod
    def from_stanford_weights(
        model_name_or_path: str | os.PathLike,
        cache_folder: str | os.PathLike | None = None,
        revision: str | None = None,
        local_files_only: bool | None = None,
        token: str | bool | None = None,
        use_auth_token: str | bool | None = None,
    ) -> "Dense":
        """Load the weight of the Dense layer using weights from a stanford-nlp checkpoint.

        Parameters
        ----------
        model_name_or_path (`str` or `os.PathLike`):
            This can be either:
            - a string, the *model id* of a model repo on huggingface.co.
            - a path to a *directory* potentially containing the file.
        cache_folder (`str` or `os.PathLike`, *optional*):
            Path to a directory in which a downloaded pretrained model configuration should be cached if the standard
            cache should not be used.
        token (`str` or *bool*, *optional*):
            The token to use as HTTP bearer authorization for remote files. If `True`, will use the token generated
            when running `huggingface-cli login` (stored in `~/.huggingface`).
        revision (`str`, *optional*, defaults to `"main"`):
            The specific model version to use. It can be a branch name, a tag name, or a commit id, since we use a
            git-based system for storing models and other artifacts on huggingface.co, so `revision` can be any
            identifier allowed by git.
        local_files_only (`bool`, *optional*, defaults to `False`):
            If `True`, will only try to load the tokenizer configuration from local files.
        """
        # Check if the model is locally available
        if not (os.path.exists(os.path.join(model_name_or_path))):
            # Else download the model/use the cached version. We first try to use the safetensors version and fall back to bin if not existing. All the recent stanford-nlp models are safetensors but we keep bin for compatibility.
            try:
                model_name_or_path = cached_file(
                    model_name_or_path,
                    filename="model.safetensors",
                    cache_dir=cache_folder,
                    revision=revision,
                    local_files_only=local_files_only,
                    token=token,
                    use_auth_token=use_auth_token,
                )
            except EnvironmentError:
                logging.warning("No safetensor model found, falling back to bin.")
                model_name_or_path = cached_file(
                    model_name_or_path,
                    filename="pytorch_model.bin",
                    cache_dir=cache_folder,
                    revision=revision,
                    local_files_only=local_files_only,
                    token=token,
                    use_auth_token=use_auth_token,
                )
        # If the model a local folder, load the safetensor
        # Again, we first try to load the safetensors version and fall back to bin if not existing.
        else:
            if os.path.exists(os.path.join(model_name_or_path, "model.safetensors")):
                model_name_or_path = os.path.join(
                    model_name_or_path, "model.safetensors"
                )
            else:
                logging.warning("No safetensor model found, falling back to bin.")
                model_name_or_path = os.path.join(
                    model_name_or_path, "pytorch_model.bin"
                )
        if model_name_or_path.endswith("safetensors"):
            with safe_open(model_name_or_path, framework="pt", device="cpu") as f:
                state_dict = {"linear.weight": f.get_tensor("linear.weight")}
        else:
            state_dict = {
                "linear.weight": torch.load(model_name_or_path, map_location="cpu")[
                    "linear.weight"
                ]
            }

        # Determine input and output dimensions
        in_features = state_dict["linear.weight"].shape[1]
        out_features = state_dict["linear.weight"].shape[0]

        # Create Dense layer instance
        model = Dense(in_features=in_features, out_features=out_features, bias=False)

        model.load_state_dict(state_dict)
        return model

    def get_config_dict(self):
        return {**super().get_config_dict(), "use_residual": self.use_residual}

    @staticmethod
    def load(input_path) -> "Dense":
        """Load a Dense layer."""
        with open(os.path.join(input_path, "config.json")) as fIn:
            config = json.load(fIn)

        config["activation_function"] = import_from_string(
            config["activation_function"]
        )()

        model = Dense(**config)

        if os.path.exists(os.path.join(input_path, "model.safetensors")):
            load_safetensors_model(model, os.path.join(input_path, "model.safetensors"))
            return model

        model.load_state_dict(
            torch.load(
                os.path.join(input_path, "pytorch_model.bin"),
                map_location=torch.device("cpu"),
            )
        )
        return model



# Could be .tsv or .json. The latter always allows more customization via optional parameters.
# I think it could be worth doing some kind of parallel reads too, if the file exceeds 1 GiBs.
# Just need to use a datastructure that shares things across processes without too much pickling.
# I think multiprocessing.Manager can do that!

import itertools

from pylate.indexes.stanford_nlp.infra.run import Run


class Collection:
    def __init__(self, path=None, data=None):
        self.path = path
        self.data = data or self._load_file(path)

    def __iter__(self):
        # TODO: If __data isn't there, stream from disk!
        return self.data.__iter__()

    def __getitem__(self, item):
        # TODO: Load from disk the first time this is called. Unless self.data is already not None.
        return self.data[item]

    def __len__(self):
        # TODO: Load here too. Basically, let's make data a property function and, on first call, either load or get __data.
        return len(self.data)

    def enumerate(self, rank):
        for _, offset, passages in self.enumerate_batches(rank=rank):
            for idx, passage in enumerate(passages):
                yield (offset + idx, passage)

    def enumerate_batches(self, rank, chunksize=None):
        assert rank is not None, "TODO: Add support for the rank=None case."

        chunksize = chunksize or self.get_chunksize()

        offset = 0
        iterator = iter(self)

        for chunk_idx, owner in enumerate(itertools.cycle(range(Run().nranks))):
            L = [line for _, line in zip(range(chunksize), iterator)]

            if len(L) > 0 and owner == rank:
                yield (chunk_idx, offset, L)

            offset += len(L)

            if len(L) < chunksize:
                return

    def get_chunksize(self):
        return min(
            25_000, 1 + len(self) // Run().nranks
        )  # 25k is great, 10k allows things to reside on GPU??

    @classmethod
    def cast(cls, obj):
        if type(obj) is str:
            return cls(path=obj)

        if type(obj) is list:
            return cls(data=obj)

        if type(obj) is cls:
            return obj

        assert False, f"obj has type {type(obj)} which is not compatible with cast()"



from __future__ import annotations

import logging
import os

import numpy as np
import torch
from sqlitedict import SqliteDict

from ..rank import RerankResult
from .base import Base
from .stanford_nlp import Indexer, IndexUpdater, Searcher
from .stanford_nlp.infra import ColBERTConfig

logger = logging.getLogger(__name__)


def reshape_embeddings(
    embeddings: np.ndarray | torch.Tensor,
) -> np.ndarray | torch.Tensor:
    """Reshape the embeddings, the PLAID index expects arrays with shape batch_size, n_tokens, embedding_size."""
    if isinstance(embeddings, np.ndarray):
        if len(embeddings.shape) == 2:
            return np.expand_dims(a=embeddings, axis=0)

    if isinstance(embeddings, torch.Tensor):
        return reshape_embeddings(embeddings=embeddings.cpu().detach().numpy())

    if isinstance(embeddings, list) and isinstance(embeddings[0], torch.Tensor):
        return [embedding.cpu().detach().numpy() for embedding in embeddings]

    return embeddings


class StanfordPLAID(Base):
    """PLAID index. The PLAID index is the most scalable type of index for multi-vector search and leverage PQ-IVF as well as custom kernel for decompression.

    Parameters
    ----------
    index_folder
        The folder where the index will be stored.
    index_name
        The name of the index.
    override
        Whether to override the collection if it already exists.
    embedding_size
        The number of dimensions of the embeddings.
    nbits
        The number of bits to use for the quantization.
    kmeans_niters
        The number of iterations to use for the k-means clustering.
    ndocs
        The number of candidate documents
    centroid_score_threshold
        The threshold scores for centroid pruning.
    ncells
        The number of cells to consider for search.
    search_batch_size
        The batch size to use when searching.
    use_triton
        Whether to use triton kernels when computing kmeans using fast-plaid. Triton kernels are faster, but yields some variance due to race condition, set to false to get 100% reproducable results. If unset, will use triton kernels if possible.

    Examples:
    --------
    from pylate import indexes, models

    index = indexes.PLAID(
        index_folder="indexes",
        index_name="plaid_colbert",
        override=True,
    )

    model = models.ColBERT(
        model_name_or_path="lightonai/GTE-ModernColBERT-v1",
    )

    documents_embeddings = model.encode(
    [
        "The nutritional profile of fruits is exceptional, containing essential vitamins, minerals, and antioxidants that support immune function. Regular consumption of diverse fruits has been linked to reduced risk of chronic diseases including heart disease, type 2 diabetes, and certain cancers. The fiber content in fruits also promotes digestive health and helps maintain healthy cholesterol levels. Nutritionists recommend consuming at least 2-3 servings of fresh fruits daily as part of a balanced diet.",
        "Recent nutritional studies have challenged traditional views on fruit consumption. While fruits contain natural sugars, their impact on blood glucose levels varies significantly based on ripeness, variety, and processing methods. The glycemic index of fruits ranges widely, with berries typically scoring lower than tropical varieties. Individuals with metabolic conditions should consider these factors when incorporating fruits into their dietary plans.",
        "Paris is the capital of France and one of the most populous cities in Europe. It is known for its art, fashion, and culture. The city is home to many famous landmarks, including the Eiffel Tower, the Louvre Museum, and Notre-Dame Cathedral. Paris is also known for its cuisine, with many world-renowned restaurants and cafes. The city has a rich history and has been a center of art and culture for centuries.",
    ],
        is_query=False,
    )

    index = index.add_documents(
        documents_ids=["1", "2", "3"], documents_embeddings=documents_embeddings
    )

    queries_embeddings = model.encode(
        ["not dangerous", "dangerous"],
        is_query=True,
    )

    index(queries_embeddings, k=30)

    """

    def __init__(
        self,
        index_folder: str = "indexes",
        index_name: str = "colbert",
        override: bool = False,
        embedding_size: int = 128,
        nbits: int = 2,
        nranks: int = 1,
        kmeans_niters: int = 4,
        index_bsize: int = 1,
        ndocs: int = 8192,
        centroid_score_threshold: float = 0.35,
        ncells: int = 8,
        search_batch_size: int = 2**18,
        use_triton: bool | None = None,
    ) -> None:
        self.config = ColBERTConfig(
            nbits=nbits,
            nranks=nranks,
            root=f"{index_folder}",
            index_root=f"{index_folder}",
            overwrite=override,
            kmeans_niters=kmeans_niters,
            index_bsize=index_bsize,
            dim=embedding_size,
            ndocs=ndocs,
            centroid_score_threshold=centroid_score_threshold,
            ncells=ncells,
            search_batch_size=search_batch_size,
            use_triton=use_triton,
        )
        self.index_name = index_name
        self.index_folder = index_folder
        self.indexer = Indexer(config=self.config)
        self.documents_ids_to_plaid_ids_path = os.path.join(
            index_folder, index_name, "documents_ids_to_plaid_ids.sqlite"
        )
        self.plaid_ids_to_documents_ids_path = os.path.join(
            index_folder, index_name, "plaid_ids_to_documents_ids.sqlite"
        )
        if not os.path.exists(index_folder):
            os.makedirs(index_folder)
        if not os.path.exists(f"{index_folder}/{index_name}"):
            os.makedirs(f"{index_folder}/{index_name}")
        if override:
            if os.path.exists(self.documents_ids_to_plaid_ids_path):
                os.remove(self.documents_ids_to_plaid_ids_path)
            if os.path.exists(self.plaid_ids_to_documents_ids_path):
                os.remove(self.plaid_ids_to_documents_ids_path)
            self.searcher = None
        else:
            documents_ids_to_plaid_ids = self._load_documents_ids_to_plaid_ids()
            # Check if the collection has already been populated and if so, load the searcher
            if len(documents_ids_to_plaid_ids) == 0:
                self.searcher = None
            else:
                self.searcher = Searcher(
                    index=self.index_name,
                    config=self.config,
                    index_root=f"{index_folder}/",
                )
            documents_ids_to_plaid_ids.close()

    def _load_documents_ids_to_plaid_ids(self) -> SqliteDict:
        """Load the SQLite database that maps document IDs to PLAID IDs."""
        return SqliteDict(self.documents_ids_to_plaid_ids_path, outer_stack=False)

    def _load_plaid_ids_to_documents_ids(self) -> SqliteDict:
        """Load the SQLite database that maps PLAID IDs to document IDs."""
        return SqliteDict(self.plaid_ids_to_documents_ids_path, outer_stack=False)

    def add_documents(
        self,
        documents_ids: str | list[str],
        documents_embeddings: list[np.ndarray | torch.Tensor],
        batch_size: int = 2000,
    ) -> None:
        """Add documents to the index."""
        if isinstance(documents_ids, str):
            documents_ids = [documents_ids]

        """Add documents to the index."""
        documents_embeddings = reshape_embeddings(embeddings=documents_embeddings)
        documents_ids_to_plaid_ids = self._load_documents_ids_to_plaid_ids()
        plaid_ids_to_documents_ids = self._load_plaid_ids_to_documents_ids()
        if self.searcher is None:
            self.indexer.index(
                name=f"{self.index_name}",
                collection=documents_embeddings,
                overwrite=True,
            )
            self.searcher = Searcher(
                index=self.index_name,
                config=self.config,
                index_root=f"{self.index_folder}/",
            )
            plaid_ids = list(range(len(documents_embeddings)))
        else:
            logger.warning(
                "You are adding documents to an existing index. This is an experimental feature and may result in suboptimal results. Please consider reindexing the entire collection (use override=True) or make sure to add as many documents during the first addition to provide the most accurate kmeans centroids."
            )
            index_updater = IndexUpdater(self.config, self.searcher)
            plaid_ids = index_updater.add(documents_embeddings)
            index_updater.persist_to_disk()

        # Get total number of existing documents
        for plaid_id, document_id in zip(plaid_ids, documents_ids):
            documents_ids_to_plaid_ids[document_id] = plaid_id
            plaid_ids_to_documents_ids[plaid_id] = document_id

        documents_ids_to_plaid_ids.commit()
        documents_ids_to_plaid_ids.close()

        plaid_ids_to_documents_ids.commit()
        plaid_ids_to_documents_ids.close()

        return self

    def remove_documents(self, documents_ids: list[str]) -> None:
        """Remove documents from the index.

        Parameters
        ----------
        documents_ids
            The documents IDs to remove.

        """
        document_ids_to_plaid_ids = self._load_documents_ids_to_plaid_ids()
        plaid_ids_to_documents_ids = self._load_plaid_ids_to_documents_ids()
        plaid_ids = [
            document_ids_to_plaid_ids[document_id] for document_id in documents_ids
        ]
        index_updater = IndexUpdater(self.config, self.searcher)
        index_updater.remove(plaid_ids)
        index_updater.persist_to_disk()
        for document_id in documents_ids:
            plaid_id = document_ids_to_plaid_ids[document_id]
            del document_ids_to_plaid_ids[document_id]
            del plaid_ids_to_documents_ids[plaid_id]

        document_ids_to_plaid_ids.commit()
        document_ids_to_plaid_ids.close()

        plaid_ids_to_documents_ids.commit()
        plaid_ids_to_documents_ids.close()

        return self

    def __call__(
        self,
        queries_embeddings: np.ndarray | torch.Tensor,
        k: int = 10,
    ) -> list[list[RerankResult]]:
        """Query the index for the nearest neighbors of the queries embeddings.

        Parameters
        ----------
        queries_embeddings
            The queries embeddings.
        k
            The number of nearest neighbors to return.

        """
        plaid_ids_to_documents_ids = self._load_plaid_ids_to_documents_ids()
        documents = []
        distances = []
        queries_embeddings = reshape_embeddings(embeddings=queries_embeddings)
        for query_embeddings in queries_embeddings:
            result = self.searcher.search(query_embeddings, k=k)
            documents.append([plaid_ids_to_documents_ids[r] for r in result[0]])
            distances.append(result[2])
        results = [
            [
                RerankResult(id=doc_id, score=score)
                for doc_id, score in zip(query_documents, query_distances)
            ]
            for query_documents, query_distances in zip(documents, distances)
        ]
        return results

    def get_documents_embeddings(
        self, document_ids: list[list[str]]
    ) -> list[list[list[int | float]]]:
        raise NotImplementedError



from __future__ import annotations

import math

import pytest
import torch

from pylate import models, rank


@pytest.mark.flaky(reruns=3, reruns_delay=5)
def test_model_creation(**kwargs) -> None:
    """Test the creation of different models."""
    query = ["fruits are healthy."]
    documents = [["fruits are healthy.", "fruits are good for health."]]
    torch.manual_seed(42)
    # Creation from a base encoder
    model = models.ColBERT(model_name_or_path="bert-base-uncased")
    # We don't test the embeddings of newly initied models for now as we need to make it deterministic
    # queries_embeddings = model.encode(sentences=query, is_query=True)
    # documents_embeddings = model.encode(sentences=documents, is_query=False)
    # reranked_documents = rank.rerank(
    #     documents_ids=[["1", "2"]],
    #     queries_embeddings=queries_embeddings,
    #     documents_embeddings=documents_embeddings,
    # )
    # assert math.isclose(
    #     reranked_documents[0][0]["score"], 25.92, rel_tol=0.01, abs_tol=0.01
    # )
    # assert math.isclose(reranked_documents[0][1]["score"], 23.7, rel_tol=0.01, abs_tol=0.01)

    # Creation from a base sentence-transformer
    model = models.ColBERT(model_name_or_path="sentence-transformers/all-MiniLM-L6-v2")
    # We don't test the embeddings of newly initied models for now as we need to make it deterministic
    # queries_embeddings = model.encode(sentences=query, is_query=True)
    # documents_embeddings = model.encode(sentences=documents, is_query=False)
    # reranked_documents = rank.rerank(
    #     documents_ids=[["1", "2"]],
    #     queries_embeddings=queries_embeddings,
    #     documents_embeddings=documents_embeddings,
    # )
    # assert math.isclose(
    #     reranked_documents[0][0]["score"], 18.77, rel_tol=0.01, abs_tol=0.01
    # )
    # assert math.isclose(
    #     reranked_documents[0][1]["score"], 18.63, rel_tol=0.01, abs_tol=0.01

    # Creation from stanford-nlp (safetensor)
    model = models.ColBERT(model_name_or_path="answerdotai/answerai-colbert-small-v1")
    queries_embeddings = model.encode(sentences=query, is_query=True)
    documents_embeddings = model.encode(sentences=documents, is_query=False)
    reranked_documents = rank.rerank(
        documents_ids=[["1", "2"]],
        queries_embeddings=queries_embeddings,
        documents_embeddings=documents_embeddings,
    )
    assert math.isclose(
        reranked_documents[0][0]["score"], 31.71, rel_tol=0.01, abs_tol=0.01
    )
    assert math.isclose(
        reranked_documents[0][1]["score"], 31.64, rel_tol=0.01, abs_tol=0.01
    )

    # Creation from stanford-nlp (bin)
    model = models.ColBERT(model_name_or_path="Crystalcareai/Colbertv2")
    queries_embeddings = model.encode(sentences=query, is_query=True)
    documents_embeddings = model.encode(sentences=documents, is_query=False)
    reranked_documents = rank.rerank(
        documents_ids=[["1", "2"]],
        queries_embeddings=queries_embeddings,
        documents_embeddings=documents_embeddings,
    )
    assert math.isclose(
        reranked_documents[0][0]["score"], 31.15, rel_tol=0.01, abs_tol=0.01
    )
    assert math.isclose(
        reranked_documents[0][1]["score"], 30.61, rel_tol=0.01, abs_tol=0.01
    )

    # Creation from PyLate
    model = models.ColBERT(model_name_or_path="lightonai/colbertv2.0")
    queries_embeddings = model.encode(sentences=query, is_query=True)
    documents_embeddings = model.encode(sentences=documents, is_query=False)
    reranked_documents = rank.rerank(
        documents_ids=[["1", "2"]],
        queries_embeddings=queries_embeddings,
        documents_embeddings=documents_embeddings,
    )
    assert math.isclose(
        reranked_documents[0][0]["score"], 30.01, rel_tol=0.01, abs_tol=0.01
    )
    assert math.isclose(
        reranked_documents[0][1]["score"], 26.98, rel_tol=0.01, abs_tol=0.01
    )



from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass, field, fields
from pathlib import Path
from platform import python_version
from typing import TYPE_CHECKING, Any, Literal

import torch
import transformers
from sentence_transformers import SentenceTransformerModelCardData
from sentence_transformers import __version__ as sentence_transformers_version
from sentence_transformers.util import (
    is_accelerate_available,
    is_datasets_available,
)
from torch import nn
from transformers.integrations import CodeCarbonCallback

from ..__version__ import __version__ as pylate_version

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from sentence_transformers.evaluation.SentenceEvaluator import SentenceEvaluator
    from sentence_transformers.SentenceTransformer import SentenceTransformer
    from sentence_transformers.trainer import SentenceTransformerTrainer


IGNORED_FIELDS = ["model", "trainer", "eval_results_dict"]


def get_versions() -> dict[str, Any]:
    versions = {
        "python": python_version(),
        "sentence_transformers": sentence_transformers_version,
        "transformers": transformers.__version__,
        "torch": torch.__version__,
        "pylate": pylate_version,
    }
    if is_accelerate_available():
        from accelerate import __version__ as accelerate_version

        versions["accelerate"] = accelerate_version
    if is_datasets_available():
        from datasets import __version__ as datasets_version

        versions["datasets"] = datasets_version
    from tokenizers import __version__ as tokenizers_version

    versions["tokenizers"] = tokenizers_version
    return versions


@dataclass
class PylateModelCardData(SentenceTransformerModelCardData):
    """
    A dataclass for storing data used in the model card.

    Parameters
    ----------
    language
        The model language, either a string or a list of strings, e.g., "en" or ["en", "de", "nl"].
    license
        The license of the model, e.g., "apache-2.0", "mit", or "cc-by-nc-sa-4.0".
    model_name
        The pretty name of the model, e.g., "SentenceTransformer based on microsoft/mpnet-base".
    model_id
        The model ID for pushing the model to the Hub, e.g., "tomaarsen/sbert-mpnet-base-allnli".
    train_datasets
        A list of dictionaries containing names and/or Hugging Face dataset IDs for training datasets,
        e.g., [{"name": "SNLI", "id": "stanfordnlp/snli"}, {"name": "MultiNLI", "id": "nyu-mll/multi_nli"}, {"name": "STSB"}].
    eval_datasets
        A list of dictionaries containing names and/or Hugging Face dataset IDs for evaluation datasets,
        e.g., [{"name": "SNLI", "id": "stanfordnlp/snli"}, {"id": "mteb/stsbenchmark-sts"}].
    task_name
        The human-readable task the model is trained on, e.g., "semantic textual similarity, semantic search, paraphrase mining, text classification, clustering, and more".
    tags
        A list of tags for the model, e.g., ["sentence-transformers", "sentence-similarity", "feature-extraction"].
    """

    # Potentially provided by the user
    language: str | list[str] | None = field(default_factory=list)
    license: str | None = None
    model_name: str | None = None
    model_id: str | None = None
    train_datasets: list[dict[str, str]] = field(default_factory=list)
    eval_datasets: list[dict[str, str]] = field(default_factory=list)
    task_name: str = "semantic textual similarity, semantic search, paraphrase mining, text classification, clustering, and more"
    tags: list[str] | None = field(
        default_factory=lambda: [
            "ColBERT",
            "PyLate",
            "sentence-transformers",
            "sentence-similarity",
            "feature-extraction",
        ]
    )
    generate_widget_examples: Literal["deprecated"] = "deprecated"

    # Automatically filled by `ModelCardCallback` and the Trainer directly
    base_model: str | None = field(default=None, init=False)
    base_model_revision: str | None = field(default=None, init=False)
    non_default_hyperparameters: dict[str, Any] = field(
        default_factory=dict, init=False
    )
    all_hyperparameters: dict[str, Any] = field(default_factory=dict, init=False)
    eval_results_dict: dict[SentenceEvaluator, dict[str, Any]] | None = field(
        default_factory=dict, init=False
    )
    training_logs: list[dict[str, float]] = field(default_factory=list, init=False)
    widget: list[dict[str, str]] = field(default_factory=list, init=False)
    predict_example: str | None = field(default=None, init=False)
    label_example_list: list[dict[str, str]] = field(default_factory=list, init=False)
    code_carbon_callback: CodeCarbonCallback | None = field(default=None, init=False)
    citations: dict[str, str] = field(default_factory=dict, init=False)
    best_model_step: int | None = field(default=None, init=False)
    trainer: SentenceTransformerTrainer | None = field(
        default=None, init=False, repr=False
    )
    datasets: list[str] = field(default_factory=list, init=False, repr=False)

    # Utility fields
    first_save: bool = field(default=True, init=False)
    widget_step: int = field(default=-1, init=False)

    # Computed once, always unchanged
    pipeline_tag: str = field(default="sentence-similarity", init=False)
    library_name: str = field(default="PyLate", init=False)
    version: dict[str, str] = field(default_factory=get_versions, init=False)
    template_path: Path = field(
        default=Path(__file__).parent / "model_card_template.md", init=False
    )

    # Passed via `register_model` only
    model: SentenceTransformer | None = field(default=None, init=False, repr=False)

    def set_losses(self, losses: list[nn.Module]) -> None:
        citations = {
            "Sentence Transformers": """
@inproceedings{reimers-2019-sentence-bert,
    title = "Sentence-BERT: Sentence Embeddings using Siamese BERT-Networks",
    author = "Reimers, Nils and Gurevych, Iryna",
    booktitle = "Proceedings of the 2019 Conference on Empirical Methods in Natural Language Processing",
    month = "11",
    year = "2019",
    publisher = "Association for Computational Linguistics",
    url = "https://arxiv.org/abs/1908.10084"
}""",
            "PyLate": """
@misc{PyLate,
title={PyLate: Flexible Training and Retrieval for Late Interaction Models},
author={Chaffin, Antoine and Sourty, Raphaël},
url={https://github.com/lightonai/pylate},
year={2024}
}""",
        }
        for loss in losses:
            try:
                citations[loss.__class__.__name__] = loss.citation
            except Exception:
                pass
        inverted_citations = defaultdict(list)
        for loss, citation in citations.items():
            inverted_citations[citation].append(loss)

        def join_list(losses: list[str]) -> str:
            if len(losses) > 1:
                return ", ".join(losses[:-1]) + " and " + losses[-1]
            return losses[0]

        self.citations = {
            join_list(losses): citation
            for citation, losses in inverted_citations.items()
        }
        self.add_tags(
            [
                f"loss:{loss}"
                for loss in {loss.__class__.__name__: loss for loss in losses}
            ]
        )

    def to_dict(self) -> dict[str, Any]:
        # Try to set the base model
        if self.first_save and not self.base_model:
            try:
                self.try_to_set_base_model()
            except Exception:
                pass

        # Set the model name
        if not self.model_name:
            if self.base_model:
                self.model_name = f"PyLate model based on {self.base_model}"
            else:
                self.model_name = "PyLate"

        super_dict = {field.name: getattr(self, field.name) for field in fields(self)}

        # Compute required formats from the (usually post-training) evaluation data
        if self.eval_results_dict:
            try:
                super_dict.update(self.format_eval_metrics())
            except Exception as exc:
                logger.warning(f"Error while formatting evaluation metrics: {exc}")
                raise exc

        # Compute required formats for the during-training evaluation data
        if self.training_logs:
            try:
                super_dict.update(self.format_training_logs())
            except Exception as exc:
                logger.warning(f"Error while formatting training logs: {exc}")

        super_dict["hide_eval_lines"] = len(self.training_logs) > 100

        # Try to add the code carbon callback data
        if (
            self.code_carbon_callback
            and self.code_carbon_callback.tracker
            and self.code_carbon_callback.tracker._start_time is not None
        ):
            super_dict.update(self.get_codecarbon_data())

        # Add some additional metadata stored in the model itself
        super_dict["document_length"] = self.model.document_length
        super_dict["query_length"] = self.model.query_length
        super_dict["output_dimensionality"] = (
            self.model.get_sentence_embedding_dimension()
        )
        super_dict["model_string"] = str(self.model)

        if self.model.similarity_fn_name:
            super_dict["similarity_fn_name"] = self.model.similarity_fn_name
        else:
            super_dict["similarity_fn_name"] = "MaxSim"
        self.first_save = False

        for key in IGNORED_FIELDS:
            super_dict.pop(key, None)
        return super_dict

    def set_widget_examples(self, dataset) -> None:
        """
        A function to create widget examples from a dataset. For now, set_widget_examples is not compatible with our transform/map operations, so we make it a no-op until it is fixed

        Parameters
        ----------
        dataset
            The dataset to create widget examples from.
        """
        pass



from dataclasses import dataclass

from .base_config import BaseConfig
from .settings import (
    DocSettings,
    IndexingSettings,
    ResourceSettings,
    RunSettings,
    SearchSettings,
)


@dataclass
class RunConfig(BaseConfig, RunSettings):
    pass


@dataclass
class ColBERTConfig(
    RunSettings,
    ResourceSettings,
    DocSettings,
    IndexingSettings,
    SearchSettings,
    BaseConfig,
):
    pass



"""Script to reproduce the training of Reason-ModernColBERT using contrastive loss on ReasonIR HQ dataset."""

from datasets import load_dataset
from sentence_transformers import (
    SentenceTransformerTrainer,
    SentenceTransformerTrainingArguments,
)

from pylate import losses, models, utils


def main():
    # As ReasonIR do not re-upload the BRIGHT data, we need to load it from the original source
    def get_doc_and_ids(doc_pairs):
        doc_ids = []
        documents = []
        for dp in doc_pairs:
            doc_ids.append(str(dp["id"]))
            documents.append(dp["content"])
        return documents, doc_ids

    def process_pos_id2doc(entry, id2doc):
        pos_docs = entry["pos"]
        neg_docs = entry["neg"]
        entry["pos"] = pos_docs[0][0] + " " + id2doc[pos_docs[0][1]]
        entry["query"] = " ".join(entry["query"])
        entry["neg"] = neg_docs[0][0] + " " + neg_docs[0][1]
        return entry

    hq_dataset = load_dataset("reasonir/reasonir-data", "hq")
    bright_docs = load_dataset("xlangai/BRIGHT", "documents")
    all_docs = []
    all_ids = []
    for task in bright_docs.keys():
        docs, ids = get_doc_and_ids(bright_docs[task])
        all_docs.extend(docs)
        all_ids.extend(ids)

    id2doc = {}
    for i in range(len(all_docs)):
        id2doc[all_ids[i]] = all_docs[i]

    train_dataset = hq_dataset.map(lambda x: process_pos_id2doc(x, id2doc))

    # Define training parameters
    num_train_epochs = 3
    lr = 1e-5
    batch_size = 256
    mini_batch_size = 32
    model_name = "lightonai/GTE-ModernColBERT-v1"
    model_shortname = model_name.split("/")[-1]

    # Set run name and output directory
    run_name = f"{model_shortname}-ReasonIR"
    output_dir = f"output/{model_shortname}/{run_name}"

    # Initialize model
    model = models.ColBERT(
        model_name_or_path=model_name,
        document_length=8192,
        query_length=128,
        skiplist_words=[],
    )

    # Setup evaluation and loss
    train_loss = losses.CachedContrastive(
        model=model,
        mini_batch_size=mini_batch_size,
        gather_across_devices=True,
        temperature=1.0,
    )

    # Configure training arguments
    args = SentenceTransformerTrainingArguments(
        output_dir=output_dir,
        num_train_epochs=num_train_epochs,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=batch_size,
        save_steps=500,
        logging_steps=1,
        fp16=False,
        bf16=True,
        run_name=run_name,
        learning_rate=lr,
        dataloader_num_workers=8,
    )

    # Initialize and run trainer
    trainer = SentenceTransformerTrainer(
        model=model,
        args=args,
        train_dataset=train_dataset,
        loss=train_loss,
        data_collator=utils.ColBERTCollator(model.tokenize),
    )

    trainer.train()
    model.save_pretrained(f"{output_dir}/final")


if __name__ == "__main__":
    main()



from __future__ import annotations

from .__version__ import __version__

__all__ = [
    "evaluation",
    "indexes",
    "losses",
    "models",
    "rank",
    "retrieve",
    "scores",
    "utils",
    "__version__",
]



import os

import torch.multiprocessing as mp

from pylate.indexes.stanford_nlp.indexing.collection_indexer import encode
from pylate.indexes.stanford_nlp.infra.launcher import Launcher
from pylate.indexes.stanford_nlp.utils.utils import create_directory, print_message


class Indexer:
    def __init__(self, config=None, verbose: int = 3):
        """
        Use Run().context() to choose the run's configuration. They are NOT extracted from `config`.
        """

        self.index_path = None
        self.verbose = verbose
        self.config = config

        # TODO: This is a hack to avoid forking in the current implementation. This should be removed in the future.
        self.config.avoid_fork_if_possible = True

    def configure(self, **kw_args):
        self.config.configure(**kw_args)

    def get_index(self):
        return self.index_path

    def erase(self, force_silent: bool = False):
        assert self.index_path is not None
        directory = self.index_path
        deleted = []

        for filename in sorted(os.listdir(directory)):
            filename = os.path.join(directory, filename)

            delete = filename.endswith(".json")
            delete = delete and (
                "metadata" in filename or "doclen" in filename or "plan" in filename
            )
            delete = delete or filename.endswith(".pt")

            if delete:
                deleted.append(filename)

        if len(deleted):
            if not force_silent:
                print_message(
                    f"#> Will delete {len(deleted)} files already at {directory}."
                )

            for filename in deleted:
                os.remove(filename)

        return deleted

    def index(self, name, collection, overwrite=False):
        assert overwrite in [True, False, "reuse", "resume", "force_silent_overwrite"]

        self.configure(
            collection=collection, index_name=name, resume=overwrite == "resume"
        )
        # Note: The bsize value set here is ignored internally. Users are encouraged
        # to supply their own batch size for indexing by using the index_bsize parameter in the ColBERTConfig.
        self.configure(bsize=64, partitions=None)

        self.index_path = self.config.index_path_
        index_does_not_exist = not os.path.exists(self.config.index_path_)

        assert (
            overwrite in [True, "reuse", "resume", "force_silent_overwrite"]
        ) or index_does_not_exist, self.config.index_path_
        create_directory(self.config.index_path_)

        if overwrite == "force_silent_overwrite":
            self.erase(force_silent=True)
        elif overwrite is True:
            self.erase()

        if index_does_not_exist or overwrite != "reuse":
            self.__launch(collection)

        return self.index_path

    def __launch(self, collection):
        launcher = Launcher(encode)
        if self.config.nranks == 1 and self.config.avoid_fork_if_possible:
            shared_queues = []
            shared_lists = []
            launcher.launch_without_fork(
                self.config, collection, shared_lists, shared_queues, self.verbose
            )

            return
        manager = mp.Manager()
        shared_lists = [manager.list() for _ in range(self.config.nranks)]
        shared_queues = [manager.Queue(maxsize=1) for _ in range(self.config.nranks)]

        # Encodes collection into index using the CollectionIndexer class
        launcher.launch(
            self.config, collection, shared_lists, shared_queues, self.verbose
        )



from __future__ import annotations

import heapq
import json
import logging
import os
from contextlib import nullcontext
from typing import TYPE_CHECKING

import torch
from sentence_transformers.evaluation import InformationRetrievalEvaluator
from torch import Tensor
from tqdm import trange

if TYPE_CHECKING:
    from ..models import ColBERT

logger = logging.getLogger(__name__)


class PyLateInformationRetrievalEvaluator(InformationRetrievalEvaluator):
    """
    This class evaluates an Information Retrieval (IR) setting. This is a direct extension of the InformationRetrievalEvaluator from the sentence-transformers library, only override the compute_metrices method to be compilatible with PyLate models (define assymetric encoding using is_query params and add padding).
    """

    def compute_metrices(
        self,
        model: ColBERT,
        corpus_model=None,
        corpus_embeddings: Tensor | None = None,
        output_path: str | None = None,
    ) -> dict[str, float]:
        if corpus_model is None:
            corpus_model = model

        max_k = max(
            max(self.mrr_at_k),
            max(self.ndcg_at_k),
            max(self.accuracy_at_k),
            max(self.precision_recall_at_k),
            max(self.map_at_k),
        )
        # Compute embedding for the queries
        with (
            nullcontext()
            if self.truncate_dim is None
            else model.truncate_sentence_embeddings(self.truncate_dim)
        ):
            query_embeddings = torch.nn.utils.rnn.pad_sequence(
                model.encode(
                    self.queries,
                    prompt_name=self.query_prompt_name,
                    prompt=self.query_prompt,
                    batch_size=self.batch_size,
                    is_query=True,
                    show_progress_bar=self.show_progress_bar,
                    convert_to_tensor=True,
                ),
                batch_first=True,
                padding_value=0,
            )

        queries_result_list = {}
        for name in self.score_functions:
            queries_result_list[name] = [[] for _ in range(len(query_embeddings))]

        # Iterate over chunks of the corpus
        for corpus_start_idx in trange(
            0,
            len(self.corpus),
            self.corpus_chunk_size,
            desc="Corpus Chunks",
            disable=not self.show_progress_bar,
        ):
            corpus_end_idx = min(
                corpus_start_idx + self.corpus_chunk_size, len(self.corpus)
            )

            # Encode chunk of corpus
            if corpus_embeddings is None:
                with (
                    nullcontext()
                    if self.truncate_dim is None
                    else corpus_model.truncate_sentence_embeddings(self.truncate_dim)
                ):
                    sub_corpus_embeddings = torch.nn.utils.rnn.pad_sequence(
                        corpus_model.encode(
                            self.corpus[corpus_start_idx:corpus_end_idx],
                            prompt_name=self.corpus_prompt_name,
                            prompt=self.corpus_prompt,
                            is_query=False,
                            batch_size=self.batch_size,
                            show_progress_bar=False,
                            # convert_to_tensor=True,
                            convert_to_numpy=False,
                        ),
                        batch_first=True,
                        padding_value=0,
                    )

            else:
                sub_corpus_embeddings = corpus_embeddings[
                    corpus_start_idx:corpus_end_idx
                ]

            # Compute cosine similarites
            for name, score_function in self.score_functions.items():
                pair_scores = score_function(query_embeddings, sub_corpus_embeddings)
                # Get top-k values
                pair_scores_top_k_values, pair_scores_top_k_idx = torch.topk(
                    pair_scores,
                    min(max_k, len(pair_scores[0])),
                    dim=1,
                    largest=True,
                    sorted=False,
                )
                pair_scores_top_k_values = pair_scores_top_k_values.cpu().tolist()
                pair_scores_top_k_idx = pair_scores_top_k_idx.cpu().tolist()

                for query_itr in range(len(query_embeddings)):
                    for sub_corpus_id, score in zip(
                        pair_scores_top_k_idx[query_itr],
                        pair_scores_top_k_values[query_itr],
                    ):
                        corpus_id = self.corpus_ids[corpus_start_idx + sub_corpus_id]
                        # NOTE: TREC/BEIR/MTEB skips cases where the corpus_id is the same as the query_id, e.g.:
                        # if corpus_id == self.queries_ids[query_itr]:
                        #     continue
                        # This is not done here, as this might be unexpected behaviour if the user just uses
                        # sets of integers from 0 as query_ids and corpus_ids.
                        if len(queries_result_list[name][query_itr]) < max_k:
                            # heaqp tracks the quantity of the first element in the tuple
                            heapq.heappush(
                                queries_result_list[name][query_itr], (score, corpus_id)
                            )
                        else:
                            heapq.heappushpop(
                                queries_result_list[name][query_itr], (score, corpus_id)
                            )

        for name in queries_result_list:
            for query_itr in range(len(queries_result_list[name])):
                for doc_itr in range(len(queries_result_list[name][query_itr])):
                    score, corpus_id = queries_result_list[name][query_itr][doc_itr]
                    queries_result_list[name][query_itr][doc_itr] = {
                        "corpus_id": corpus_id,
                        "score": score,
                    }

        if self.write_predictions and output_path is not None:
            for name in queries_result_list:
                base_filename = self.predictions_file.replace(
                    ".jsonl", f"_{name}.jsonl"
                )
                json_path = os.path.join(output_path, base_filename)
                mode = "w"  # Always create a new file for each score function

                with open(json_path, mode=mode, encoding="utf-8") as fOut:
                    for query_itr in range(len(queries_result_list[name])):
                        query_id = self.queries_ids[query_itr]
                        query_text = self.queries[query_itr]
                        results = queries_result_list[name][query_itr]

                        # Sort results by score in descending order
                        results = sorted(
                            results, key=lambda x: x["score"], reverse=True
                        )

                        prediction = {
                            "query_id": query_id,
                            "query": query_text,
                            "results": results,
                        }

                        fOut.write(json.dumps(prediction) + "\n")

        logger.info(f"Queries: {len(self.queries)}")
        logger.info(f"Corpus: {len(self.corpus)}\n")

        # Compute scores
        scores = {
            name: self.compute_metrics(queries_result_list[name])
            for name in self.score_functions
        }

        # Output
        for name in self.score_function_names:
            logger.info(f"Score-Function: {name}")
            self.output_scores(scores[name])

        return scores



from __future__ import annotations

import os

import torch

from pylate.indexes.stanford_nlp.infra.config import ColBERTConfig
from pylate.indexes.stanford_nlp.infra.launcher import print_memory_stats
from pylate.indexes.stanford_nlp.infra.run import Run
from pylate.indexes.stanford_nlp.search.index_storage import IndexScorer


class Searcher:
    def __init__(
        self,
        index,
        checkpoint=None,
        collection=None,
        config=None,
        index_root=None,
        verbose: int = 3,
    ):
        self.verbose = verbose
        if self.verbose > 1:
            print_memory_stats()

        initial_config = ColBERTConfig.from_existing(config, Run().config)

        default_index_root = initial_config.index_root_
        index_root = index_root if index_root else default_index_root
        self.index = os.path.join(index_root, index)
        self.index_config = ColBERTConfig.load_from_index(self.index)

        self.config = ColBERTConfig.from_existing(self.index_config, initial_config)
        self.configure()

        self.use_gpu = self.config.total_visible_gpus > 0

        load_index_with_mmap = self.config.load_index_with_mmap
        if load_index_with_mmap and self.use_gpu:
            raise ValueError("Memory-mapped index can only be used with CPU!")
        self.ranker = IndexScorer(self.index, self.use_gpu, load_index_with_mmap)

        print_memory_stats()

    def configure(self, **kw_args):
        self.config.configure(**kw_args)

    def search(self, Q, k=10, filter_fn=None, full_length_search=False, pids=None):
        # print("Q", Q)
        # print("Q TYPE", Q[0].dtype)
        Q = torch.tensor(
            Q, dtype=torch.float16 if self.use_gpu else torch.float32
        ).unsqueeze(0)
        # Cast to bf16
        # Q = Q.to(torch.float16)
        # print("Q TYPE", Q.type())
        return self.dense_search(Q, k, filter_fn=filter_fn, pids=pids)

    def dense_search(self, Q: torch.Tensor, k=10, filter_fn=None, pids=None):
        if k <= 10:
            if self.config.ncells is None:
                self.configure(ncells=1)
            if self.config.centroid_score_threshold is None:
                self.configure(centroid_score_threshold=0.5)
            if self.config.ndocs is None:
                self.configure(ndocs=256)
        elif k <= 100:
            if self.config.ncells is None:
                self.configure(ncells=2)
            if self.config.centroid_score_threshold is None:
                self.configure(centroid_score_threshold=0.45)
            if self.config.ndocs is None:
                self.configure(ndocs=1024)
        else:
            if self.config.ncells is None:
                self.configure(ncells=4)
            if self.config.centroid_score_threshold is None:
                self.configure(centroid_score_threshold=0.4)
            if self.config.ndocs is None:
                self.configure(ndocs=max(k * 4, 8192))

        pids, scores = self.ranker.rank(self.config, Q, filter_fn=filter_fn, pids=pids)

        return pids[:k], list(range(1, k + 1)), scores[:k]



"""This script is responsible for building the API reference. The API reference is located in
docs/api. The script scans through all the modules, classes, and functions. It processes
the __doc__ of each object and formats it so that MkDocs can process it in turn.
"""

from __future__ import annotations

import functools
import importlib
import inspect
import os
import pathlib
import re
import shutil

from numpydoc.docscrape import ClassDoc, FunctionDoc

package = "pylate"

shutil.copy("README.md", "docs/index.md")


with open("docs/index.md", mode="r") as file:
    content = file.read()

with open("docs/index.md", mode="w") as file:
    file.write(content)


def paragraph(text):
    return f"{text}\n"


def h1(text):
    return paragraph(f"# {text}")


def h2(text):
    return paragraph(f"## {text}")


def h3(text):
    return paragraph(f"### {text}")


def h4(text):
    return paragraph(f"#### {text}")


def link(caption, href):
    return f"[{caption}]({href})"


def code(text):
    return f"`{text}`"


def li(text):
    return f"- {text}\n"


def snake_to_kebab(text):
    return text.replace("_", "-")


def inherit_docstring(c, meth):
    """Since Python 3.5, inspect.getdoc is supposed to return the docstring from a parent class
    if a class has none. However this doesn't seem to work for Cython classes.
    """

    doc = None

    for ancestor in inspect.getmro(c):
        try:
            ancestor_meth = getattr(ancestor, meth)
        except AttributeError:
            break
        doc = inspect.getdoc(ancestor_meth)
        if doc:
            break

    return doc


def inherit_signature(c, method_name):
    m = getattr(c, method_name)
    try:
        sig = inspect.signature(m)
    except ValueError:
        return inspect.Signature()

    params = []

    for param in sig.parameters.values():
        if param.name == "self" or param.annotation is not param.empty:
            params.append(param)
            continue

        for ancestor in inspect.getmro(c):
            try:
                ancestor_meth = inspect.signature(getattr(ancestor, m.__name__))
            except AttributeError:
                break
            try:
                ancestor_param = ancestor_meth.parameters[param.name]
            except KeyError:
                break
            if ancestor_param.annotation is not param.empty:
                param = param.replace(annotation=ancestor_param.annotation)
                break

        params.append(param)

    return_annotation = sig.return_annotation
    if return_annotation is inspect._empty:
        for ancestor in inspect.getmro(c):
            try:
                ancestor_meth = inspect.signature(getattr(ancestor, m.__name__))
            except AttributeError:
                break
            if ancestor_meth.return_annotation is not inspect._empty:
                return_annotation = ancestor_meth.return_annotation
                break

    return sig.replace(parameters=params, return_annotation=return_annotation)


def pascal_to_kebab(string):
    string = re.sub("(.)([A-Z][a-z]+)", r"\1-\2", string)
    string = re.sub("(.)([0-9]+)", r"\1-\2", string)
    return re.sub("([a-z0-9])([A-Z])", r"\1-\2", string).lower()


class Linkifier:
    def __init__(self):
        path_index = {}
        name_index = {}

        modules = {
            module: importlib.import_module(f"{package}.{module}")
            for module in importlib.import_module(f"{package}").__all__
        }

        def index_module(mod_name, mod, path):
            path = os.path.join(path, mod_name)
            dotted_path = path.replace("/", ".")

            for func_name, func in inspect.getmembers(mod, inspect.isfunction):
                for e in (
                    f"{mod_name}.{func_name}",
                    f"{dotted_path}.{func_name}",
                    f"{func.__module__}.{func_name}",
                ):
                    path_index[e] = os.path.join(path, snake_to_kebab(func_name))
                    name_index[e] = f"{dotted_path}.{func_name}"

            for klass_name, klass in inspect.getmembers(mod, inspect.isclass):
                for e in (
                    f"{mod_name}.{klass_name}",
                    f"{dotted_path}.{klass_name}",
                    f"{klass.__module__}.{klass_name}",
                ):
                    path_index[e] = os.path.join(path, klass_name)
                    name_index[e] = f"{dotted_path}.{klass_name}"

            for submod_name, submod in inspect.getmembers(mod, inspect.ismodule):
                if submod_name not in mod.__all__ or submod_name == "typing":
                    continue
                for e in (f"{mod_name}.{submod_name}", f"{dotted_path}.{submod_name}"):
                    path_index[e] = os.path.join(path, snake_to_kebab(submod_name))

                # Recurse
                index_module(submod_name, submod, path=path)

        for mod_name, mod in modules.items():
            index_module(mod_name, mod, path="")

        # Prepend {package} to each index entry
        for k in list(path_index.keys()):
            path_index[f"{package}.{k}"] = path_index[k]
        for k in list(name_index.keys()):
            name_index[f"{package}.{k}"] = name_index[k]

        self.path_index = path_index
        self.name_index = name_index

    def linkify(self, text, use_fences, depth):
        path = self.path_index.get(text)
        name = self.name_index.get(text)
        if path and name:
            backwards = "../" * (depth + 1)
            if use_fences:
                return f"[`{name}`]({backwards}{path})"
            return f"[{name}]({backwards}{path})"
        return None

    def linkify_fences(self, text, depth):
        between_fences = re.compile("`[\w\.]+\.\w+`")
        return between_fences.sub(
            lambda x: self.linkify(x.group().strip("`"), True, depth) or x.group(), text
        )

    def linkify_dotted(self, text, depth):
        dotted = re.compile("\w+\.[\.\w]+")
        return dotted.sub(
            lambda x: self.linkify(x.group(), False, depth) or x.group(), text
        )


def concat_lines(lines):
    return inspect.cleandoc(" ".join("\n\n" if line == "" else line for line in lines))


def print_docstring(obj, file, depth):
    """Prints a classes's docstring to a file."""

    doc = ClassDoc(obj) if inspect.isclass(obj) else FunctionDoc(obj)

    printf = functools.partial(print, file=file)

    printf(h1(obj.__name__))
    printf(linkifier.linkify_fences(paragraph(concat_lines(doc["Summary"])), depth))
    printf(
        linkifier.linkify_fences(
            paragraph(concat_lines(doc["Extended Summary"])), depth
        )
    )

    # We infer the type annotations from the signatures, and therefore rely on the signature
    # instead of the docstring for documenting parameters
    try:
        signature = inspect.signature(obj)
    except ValueError:
        signature = (
            inspect.Signature()
        )  # TODO: this is necessary for Cython classes, but it's not correct
    params_desc = {param.name: " ".join(param.desc) for param in doc["Parameters"]}

    # Parameters
    if signature.parameters:
        printf(h2("Parameters"))
    for param in signature.parameters.values():
        # Name
        printf(f"- **{param.name}**", end="")
        # Type annotation
        if param.annotation is not param.empty:
            anno = inspect.formatannotation(param.annotation)
            anno = linkifier.linkify_dotted(anno, depth)
            printf(f" (*{anno}*)", end="")
        # Default value
        if param.default is not param.empty:
            printf(f" – defaults to `{param.default}`", end="")
        printf("\n", file=file)
        # Description
        if param.name in params_desc:
            desc = params_desc[param.name]
            if desc:
                printf(f"    {desc}\n")
    printf("")

    # Attributes
    if doc["Attributes"]:
        printf(h2("Attributes"))
    for attr in doc["Attributes"]:
        # Name
        printf(f"- **{attr.name}**", end="")
        # Type annotation
        if attr.type:
            printf(f" (*{attr.type}*)", end="")
        printf("\n", file=file)
        # Description
        desc = " ".join(attr.desc)
        if desc:
            printf(f"    {desc}\n")
    printf("")

    # Examples
    if doc["Examples"]:
        printf(h2("Examples"))

        in_code = False
        after_space = False

        for line in inspect.cleandoc("\n".join(doc["Examples"])).splitlines():
            if (
                in_code
                and after_space
                and line
                and not line.startswith(">>>")
                and not line.startswith("...")
            ):
                printf("```\n")
                in_code = False
                after_space = False

            if not in_code and line.startswith(">>>"):
                printf("```python")
                in_code = True

            after_space = False
            if not line:
                after_space = True

            printf(line)

        if in_code:
            printf("```")
    printf("")

    # Methods
    if inspect.isclass(obj) and doc["Methods"]:
        printf(h2("Methods"))
        printf_indent = lambda x, **kwargs: printf(f"    {x}", **kwargs)  # noqa: E731

        for meth in doc["Methods"]:
            printf(paragraph(f'???- note "{meth.name}"'))

            # Parse method docstring
            docstring = inherit_docstring(c=obj, meth=meth.name)
            if not docstring:
                continue
            meth_doc = FunctionDoc(func=None, doc=docstring)

            printf_indent(paragraph(" ".join(meth_doc["Summary"])))
            if meth_doc["Extended Summary"]:
                printf_indent(paragraph(" ".join(meth_doc["Extended Summary"])))

            # We infer the type annotations from the signatures, and therefore rely on the signature
            # instead of the docstring for documenting parameters
            signature = inherit_signature(obj, meth.name)
            params_desc = {
                param.name: " ".join(param.desc) for param in doc["Parameters"]
            }

            # Parameters
            if (
                len(signature.parameters) > 1
            ):  # signature is never empty, but self doesn't count
                printf_indent("**Parameters**\n")
            for param in signature.parameters.values():
                if param.name == "self":
                    continue
                # Name
                printf_indent(f"- **{param.name}**", end="")
                # Type annotation
                if param.annotation is not param.empty:
                    printf_indent(
                        f" (*{inspect.formatannotation(param.annotation)}*)", end=""
                    )
                # Default value
                if param.default is not param.empty:
                    printf_indent(f" – defaults to `{param.default}`", end="")
                printf_indent("", file=file)
                # Description
                desc = params_desc.get(param.name)
                if desc:
                    printf_indent(f"    {desc}")
            printf_indent("")

            # Returns
            if meth_doc["Returns"]:
                printf_indent("**Returns**\n")
                return_val = meth_doc["Returns"][0]
                if signature.return_annotation is not inspect._empty:
                    if inspect.isclass(signature.return_annotation):
                        printf_indent(
                            f"*{signature.return_annotation.__name__}*: ", end=""
                        )
                    else:
                        printf_indent(f"*{signature.return_annotation}*: ", end="")
                printf_indent(return_val.type)
                printf_indent("")

    # Notes
    if doc["Notes"]:
        printf(h2("Notes"))
        printf(paragraph("\n".join(doc["Notes"])))

    # References
    if doc["References"]:
        printf(h2("References"))
        printf(paragraph("\n".join(doc["References"])))


def print_module(mod, path, overview, is_submodule=False):
    mod_name = mod.__name__.split(".")[-1]

    # Create a directory for the module
    mod_slug = snake_to_kebab(mod_name)
    mod_path = path.joinpath(mod_slug)
    mod_short_path = str(mod_path).replace("docs/api/", "")
    os.makedirs(mod_path, exist_ok=True)
    with open(mod_path.joinpath(".pages"), "w") as f:
        f.write(f"title: {mod_name}")

    # Add the module to the overview
    if is_submodule:
        print(h3(mod_name), file=overview)
    else:
        print(h2(mod_name), file=overview)
    if mod.__doc__:
        print(paragraph(mod.__doc__), file=overview)

    # Extract all public classes and functions
    ispublic = lambda x: x.__name__ in mod.__all__ and not x.__name__.startswith("_")  # noqa: E731
    classes = inspect.getmembers(mod, lambda x: inspect.isclass(x) and ispublic(x))
    funcs = inspect.getmembers(mod, lambda x: inspect.isfunction(x) and ispublic(x))

    # Classes

    if classes and funcs:
        print("\n**Classes**\n", file=overview)

    for _, c in classes:
        print(f"{mod_name}.{c.__name__}")

        # Add the class to the overview
        slug = snake_to_kebab(c.__name__)
        print(
            li(link(c.__name__, f"../{mod_short_path}/{slug}")), end="", file=overview
        )

        # Write down the class' docstring
        with open(mod_path.joinpath(slug).with_suffix(".md"), "w") as file:
            print_docstring(obj=c, file=file, depth=mod_short_path.count("/") + 1)

    # Functions

    if classes and funcs:
        print("\n**Functions**\n", file=overview)

    for _, f in funcs:
        print(f"{mod_name}.{f.__name__}")

        # Add the function to the overview
        slug = snake_to_kebab(f.__name__)
        print(
            li(link(f.__name__, f"../{mod_short_path}/{slug}")), end="", file=overview
        )

        # Write down the function' docstring
        with open(mod_path.joinpath(slug).with_suffix(".md"), "w") as file:
            print_docstring(obj=f, file=file, depth=mod_short_path.count(".") + 1)

    # Sub-modules
    for name, submod in inspect.getmembers(mod, inspect.ismodule):
        # We only want to go through the public submodules, such as optim.schedulers
        if (
            name in ("tags", "typing", "inspect", "skmultiflow_utils")
            or name not in mod.__all__
            or name.startswith("_")
        ):
            continue
        print_module(mod=submod, path=mod_path, overview=overview, is_submodule=True)

    print("", file=overview)


if __name__ == "__main__":
    api_path = pathlib.Path("docs/api")

    # Create a directory for the API reference
    shutil.rmtree(api_path, ignore_errors=True)
    os.makedirs(api_path, exist_ok=True)
    with open(api_path.joinpath(".pages"), "w") as f:
        f.write("title: API reference\narrange:\n  - overview.md\n  - ...\n")

    overview = open(api_path.joinpath("overview.md"), "w")
    print(h1("Overview"), file=overview)

    linkifier = Linkifier()

    for mod_name, mod in inspect.getmembers(
        importlib.import_module(f"{package}"), inspect.ismodule
    ):
        if mod_name.startswith("_"):
            continue
        print(mod_name)
        print_module(mod, path=api_path, overview=overview)



from __future__ import annotations

from enum import Enum
from typing import Callable

from numpy import ndarray
from torch import Tensor

from ..scores import colbert_scores, colbert_scores_pairwise


class SimilarityFunction(Enum):
    """
    Enum class for supported score functions. The following functions are supported:
    - ``SimilarityFunction.MAXSIM`` (``"MaxSim"``): Max similarity
    """

    MAXSIM = "MaxSim"

    @staticmethod
    def to_similarity_fn(
        similarity_function: str | SimilarityFunction,
    ) -> Callable[[Tensor | ndarray, Tensor | ndarray], Tensor]:
        """
        Converts a similarity function name or enum value to the corresponding similarity function.

        Parameters
        ----------
        similarity_function
            The name or enum value of the similarity function.
        """
        similarity_function = SimilarityFunction(similarity_function)
        if similarity_function == SimilarityFunction.MAXSIM:
            return colbert_scores

        raise ValueError(
            f"The provided function {similarity_function} is not supported. Use one of the supported values: {SimilarityFunction.possible_values()}."
        )

    @staticmethod
    def to_similarity_pairwise_fn(
        similarity_function: str | SimilarityFunction,
    ) -> Callable[[Tensor | ndarray, Tensor | ndarray], Tensor]:
        """
        Converts a similarity function into a pairwise similarity function.

        The pairwise similarity function returns the diagonal vector from the similarity matrix, i.e. it only
        computes the similarity(a[i], b[i]) for each i in the range of the input tensors, rather than
        computing the similarity between all pairs of a and b.

        Parameters
        ----------
        similarity_function
            The name or enum value of the similarity function.
        """
        similarity_function = SimilarityFunction(similarity_function)
        if similarity_function == SimilarityFunction.MAXSIM:
            return colbert_scores_pairwise

    @staticmethod
    def possible_values() -> list[str]:
        """
        Returns a list of possible values for the SimilarityFunction enum.

        Examples
        --------

        >>> possible_values = SimilarityFunction.possible_values()
        >>> possible_values
        ['MaxSim']
        """
        return [m.value for m in SimilarityFunction]



from __future__ import annotations

import logging
from typing import Literal

from sentence_transformers.evaluation.NanoBEIREvaluator import (
    NanoBEIREvaluator as NanoBEIREvaluatorST,
)
from sentence_transformers.util import is_datasets_available

from .pylate_information_retrieval_evaluator import PyLateInformationRetrievalEvaluator

logger = logging.getLogger(__name__)

DatasetNameType = Literal[
    "climatefever",
    "dbpedia",
    "fever",
    "fiqa2018",
    "hotpotqa",
    "msmarco",
    "nfcorpus",
    "nq",
    "quoraretrieval",
    "scidocs",
    "arguana",
    "scifact",
    "touche2020",
]


MAPPING_DATASET_NAME_TO_ID = {
    "climatefever": "zeta-alpha-ai/NanoClimateFEVER",
    "dbpedia": "zeta-alpha-ai/NanoDBPedia",
    "fever": "zeta-alpha-ai/NanoFEVER",
    "fiqa2018": "zeta-alpha-ai/NanoFiQA2018",
    "hotpotqa": "zeta-alpha-ai/NanoHotpotQA",
    "msmarco": "zeta-alpha-ai/NanoMSMARCO",
    "nfcorpus": "zeta-alpha-ai/NanoNFCorpus",
    "nq": "zeta-alpha-ai/NanoNQ",
    "quoraretrieval": "zeta-alpha-ai/NanoQuoraRetrieval",
    "scidocs": "zeta-alpha-ai/NanoSCIDOCS",
    "arguana": "zeta-alpha-ai/NanoArguAna",
    "scifact": "zeta-alpha-ai/NanoSciFact",
    "touche2020": "zeta-alpha-ai/NanoTouche2020",
}

MAPPING_DATASET_NAME_TO_HUMAN_READABLE = {
    "climatefever": "ClimateFEVER",
    "dbpedia": "DBPedia",
    "fever": "FEVER",
    "fiqa2018": "FiQA2018",
    "hotpotqa": "HotpotQA",
    "msmarco": "MSMARCO",
    "nfcorpus": "NFCorpus",
    "nq": "NQ",
    "quoraretrieval": "QuoraRetrieval",
    "scidocs": "SCIDOCS",
    "arguana": "ArguAna",
    "scifact": "SciFact",
    "touche2020": "Touche2020",
}


class NanoBEIREvaluator(NanoBEIREvaluatorST):
    """Evaluate the performance of a PyLate Model on the NanoBEIR collection.


    This is a direct extension of the NanoBEIREvaluator from the
    sentence-transformers library, leveraging the
    PyLateInformationRetrievalEvaluator class. The collection is a set of datasets
    based on the BEIR collection, but with a significantly smaller size, so it
    can be used for quickly evaluating the retrieval performance of a
    model before commiting to a full evaluation.
    The Evaluator will return the same metrics as the InformationRetrievalEvaluator
    (i.e., MRR, nDCG, Recall@k), for each dataset and on average.

    Examples
    --------
    >>> from pylate import models, evaluation

    >>> model = models.ColBERT(
    ...     model_name_or_path="lightonai/colbertv2.0"
    ... )

    >>> datasets = ["SciFact"]

    >>> evaluator = evaluation.NanoBEIREvaluator(
    ...    dataset_names=datasets
    ... )

    evaluator(model)
    {'NanoSciFact_MaxSim_accuracy@1': 0.62, 'NanoSciFact_MaxSim_accuracy@3': 0.74, 'NanoSciFact_MaxSim_accuracy@5': 0.8, 'NanoSciFact_MaxSim_accuracy@10': 0.86, 'NanoSciFact_MaxSim_precision@1': 0.62, 'NanoSciFact_MaxSim_precision@3': 0.26666666666666666, 'NanoSciFact_MaxSim_precision@5': 0.18, 'NanoSciFact_MaxSim_precision@10': 0.096, 'NanoSciFact_MaxSim_recall@1': 0.595, 'NanoSciFact_MaxSim_recall@3': 0.715, 'NanoSciFact_MaxSim_recall@5': 0.79, 'NanoSciFact_MaxSim_recall@10': 0.85, 'NanoSciFact_MaxSim_ndcg@10': 0.7279903941189909, 'NanoSciFact_MaxSim_mrr@10': 0.6912222222222222, 'NanoSciFact_MaxSim_map@100': 0.6903374780806633, 'NanoBEIR_mean_MaxSim_accuracy@1': 0.62, 'NanoBEIR_mean_MaxSim_accuracy@3': 0.74, 'NanoBEIR_mean_MaxSim_accuracy@5': 0.8, 'NanoBEIR_mean_MaxSim_accuracy@10': 0.86, 'NanoBEIR_mean_MaxSim_precision@1': 0.62, 'NanoBEIR_mean_MaxSim_precision@3': 0.26666666666666666, 'NanoBEIR_mean_MaxSim_precision@5': 0.18, 'NanoBEIR_mean_MaxSim_precision@10': 0.096, 'NanoBEIR_mean_MaxSim_recall@1': 0.595, 'NanoBEIR_mean_MaxSim_recall@3': 0.715, 'NanoBEIR_mean_MaxSim_recall@5': 0.79, 'NanoBEIR_mean_MaxSim_recall@10': 0.85, 'NanoBEIR_mean_MaxSim_ndcg@10': 0.7279903941189909, 'NanoBEIR_mean_MaxSim_mrr@10': 0.6912222222222222, 'NanoBEIR_mean_MaxSim_map@100': 0.6903374780806633}

    References
    ----------
    - [NanoBEIR](https://huggingface.co/collections/zeta-alpha-ai/nanobeir-66e1a0af21dfd93e620cd9f6)

    """

    def _load_dataset(
        self, dataset_name: DatasetNameType, **ir_evaluator_kwargs
    ) -> PyLateInformationRetrievalEvaluator:
        if not is_datasets_available():
            raise ValueError(
                "datasets is not available. Please install it to use the NanoBEIREvaluator."
            )
        from datasets import load_dataset

        dataset_path = MAPPING_DATASET_NAME_TO_ID[dataset_name.lower()]
        corpus = load_dataset(dataset_path, "corpus", split="train")
        queries = load_dataset(dataset_path, "queries", split="train")
        qrels = load_dataset(dataset_path, "qrels", split="train")
        corpus_dict = {
            sample["_id"]: sample["text"]
            for sample in corpus
            if len(sample["text"]) > 0
        }
        queries_dict = {
            sample["_id"]: sample["text"]
            for sample in queries
            if len(sample["text"]) > 0
        }
        qrels_dict = {}
        for sample in qrels:
            if sample["query-id"] not in qrels_dict:
                qrels_dict[sample["query-id"]] = set()
            qrels_dict[sample["query-id"]].add(sample["corpus-id"])

        if self.query_prompts is not None:
            ir_evaluator_kwargs["query_prompt"] = self.query_prompts.get(
                dataset_name, None
            )
        if self.corpus_prompts is not None:
            ir_evaluator_kwargs["corpus_prompt"] = self.corpus_prompts.get(
                dataset_name, None
            )
        human_readable_name = self._get_human_readable_name(dataset_name)
        return PyLateInformationRetrievalEvaluator(
            queries=queries_dict,
            corpus=corpus_dict,
            relevant_docs=qrels_dict,
            name=human_readable_name,
            **ir_evaluator_kwargs,
        )



from __future__ import annotations

from collections.abc import Iterator
from contextlib import nullcontext
from functools import partial
from typing import Callable, Iterable, Optional

import torch
import torch.nn.functional as F
import tqdm
from torch import Tensor, nn
from torch.utils.checkpoint import get_device_states, set_device_states

from ..models import ColBERT
from ..scores import colbert_scores
from ..utils import all_gather, all_gather_with_gradients, get_rank, get_world_size
from .contrastive import extract_skiplist_mask


class RandContext:
    """Random-state context manager class. Reference: https://github.com/luyug/GradCache.

    This class will back up the pytorch's random state during initialization. Then when the context is activated,
    the class will set up the random state with the backed-up one.
    """

    def __init__(self, *tensors) -> None:
        self.fwd_cpu_state = torch.get_rng_state()
        if torch.backends.mps.is_available():
            raise RuntimeError(
                "MPS backend is not supported for this operation. Please use CPU or CUDA."
            )
        else:
            self.fwd_gpu_devices, self.fwd_gpu_states = get_device_states(*tensors)

    def __enter__(self) -> None:
        self._fork = torch.random.fork_rng(devices=self.fwd_gpu_devices, enabled=True)
        self._fork.__enter__()
        torch.set_rng_state(self.fwd_cpu_state)
        set_device_states(self.fwd_gpu_devices, self.fwd_gpu_states)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._fork.__exit__(exc_type, exc_val, exc_tb)
        self._fork = None


def _backward_hook(
    grad_output: Tensor,
    sentence_features: Iterable[dict[str, Tensor]],
    loss_obj,
) -> None:
    """A backward hook that re-runs the forward for each mini-batch with gradients enabled
    and uses the cached partial derivatives w.r.t. the embeddings to backprop.
    """
    assert loss_obj.cache is not None
    assert loss_obj.random_states is not None
    with torch.enable_grad():
        for sentence_feature, grad, random_states in zip(
            sentence_features, loss_obj.cache, loss_obj.random_states
        ):
            for (reps_mb, _), grad_mb in zip(
                loss_obj.embed_minibatch_iter(
                    sentence_feature=sentence_feature,
                    with_grad=True,
                    copy_random_state=False,
                    random_states=random_states,
                ),
                grad,
            ):
                # Plug back the cached gradients into the backward pass by dotting them with the corresponding representations
                # Scale by grad_output from the top-level backward pass to account for gradient of downstream operations (not useful in this setup)
                surrogate = (
                    torch.dot(reps_mb.flatten(), grad_mb.flatten()) * grad_output
                )
                surrogate.backward()


class CachedContrastive(nn.Module):
    """A cached, in-batch negatives contrastive loss for PyLate, analogous to
    SentenceTransformers' CachedMultipleNegativesRankingLoss. This allows
    large effective batch sizes by chunking the embeddings pass and caching
    gradients w.r.t. those embeddings.

    Parameters
    ----------
    model :
        A PyLate ColBERT model
    score_metric
        ColBERT scoring function. Defaults to colbert_scores.
    mini_batch_size
        Chunk size for the forward pass. You can keep this small to avoid OOM on large batch sizes.
    size_average
        Whether to average or sum the cross-entropy loss across the mini-batch.
    gather_across_devices
        Whether to gather the embeddings across devices to have more in batch negatives. We recommand making sure the sampling across GPUs use the same dataset in case of multi-dataset training to make sure the negatives are plausible.
    show_progress_bar
        Whether to show a TQDM progress bar for the embedding steps.

    Examples
    --------
    >>> from pylate import models, losses

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2", device="cpu"
    ... )

    >>> loss = losses.CachedContrastive(model=model, mini_batch_size=1)

    >>> anchors = model.tokenize([
    ...     "fruits are healthy.", "chips are not healthy."
    ... ], is_query=True)

    >>> positives = model.tokenize([
    ...     "fruits are good for health.", "chips are not good for health."
    ... ], is_query=False)

    >>> negatives = model.tokenize([
    ...     "fruits are bad for health.", "chips are good for health."
    ... ], is_query=False)

    >>> sentence_features = [anchors, positives, negatives]

    >>> loss = loss(sentence_features=sentence_features)
    >>> assert isinstance(loss.item(), float)
    """

    def __init__(
        self,
        model: ColBERT,
        score_metric: Callable = colbert_scores,
        mini_batch_size: int = 32,
        size_average: bool = True,
        gather_across_devices: bool = False,
        show_progress_bar: bool = False,
        temperature: float = 1.0,
    ) -> None:
        super(CachedContrastive, self).__init__()
        self.model = model
        self.score_metric = score_metric
        self.mini_batch_size = mini_batch_size
        self.size_average = size_average
        self.gather_across_devices = gather_across_devices
        self.show_progress_bar = show_progress_bar
        self.temperature = temperature

        # Will hold partial derivatives for each embedding chunk
        self.cache: list[list[Tensor]] | None = None
        # Will hold random states for each chunk, so we can re-run the embedding pass with grads
        self.random_states: list[list[RandContext]] | None = None

    def embed_minibatch(
        self,
        sentence_feature: dict[str, Tensor],
        begin: int,
        end: int,
        with_grad: bool,
        copy_random_state: bool,
        random_state: RandContext | None = None,
    ) -> tuple[Tensor, RandContext | None]:
        """Forward pass on a slice [begin:end] of sentence_feature. If 'with_grad' is False,
        we run under torch.no_grad. If 'copy_random_state' is True, we create and return
        a RandContext so that we can exactly reproduce this forward pass later.
        """
        grad_context = nullcontext if with_grad else torch.no_grad
        random_state_context = nullcontext() if random_state is None else random_state
        sentence_feature_minibatch = {
            k: v[begin:end] for k, v in sentence_feature.items()
        }
        with random_state_context:
            with grad_context():
                # If we need a new random-state copy, create it
                random_state = (
                    RandContext(*sentence_feature_minibatch.values())
                    if copy_random_state
                    else None
                )
                outputs = self.model(sentence_feature_minibatch)
                # by default, PyLate ColBERT forward returns a dict with "token_embeddings"
                embeddings = F.normalize(outputs["token_embeddings"], p=2, dim=-1)

        return embeddings, random_state

    def embed_minibatch_iter(
        self,
        sentence_feature: dict[str, Tensor],
        with_grad: bool,
        copy_random_state: bool,
        random_states: list[RandContext] | None = None,
    ) -> Iterator[tuple[Tensor, RandContext | None]]:
        """Yields chunks of embeddings (and corresponding RandContext) for the given
        sentence_feature, respecting the mini_batch_size limit.
        """
        input_ids = sentence_feature["input_ids"]
        bsz = input_ids.size(0)
        for i, b in enumerate(
            tqdm.trange(
                0,
                bsz,
                self.mini_batch_size,
                desc="Embed mini-batches",
                disable=not self.show_progress_bar,
            )
        ):
            e = b + self.mini_batch_size
            reps, random_state = self.embed_minibatch(
                sentence_feature=sentence_feature,
                begin=b,
                end=e,
                with_grad=with_grad,
                copy_random_state=copy_random_state,
                random_state=None if random_states is None else random_states[i],
            )
            yield reps, random_state  # reps: (mbsz, hdim)

    def calculate_loss_and_cache_gradients(self, reps, masks) -> Tensor:
        """Calculate the cross-entropy loss and cache the gradients wrt. the embeddings."""
        # we want partial grads on all the chunked embeddings
        loss = self.calculate_loss(reps, masks, with_backward=True)
        loss = loss.detach().requires_grad_()

        self.cache = [
            [r.grad for r in rs] for rs in reps
        ]  # e.g. 3 * bsz/mbsz * (mbsz, hdim)

        return loss

    def calculate_loss(self, reps, masks, with_backward: bool = False) -> Tensor:
        """Calculate the cross-entropy loss. No need to cache the gradients. Each sub-list in reps is a list of mini-batch chunk embeddings

        Parameters
        ----------
        reps :
            A list of list of mini-batch chunk embeddings. The first list are the anchors, the second are the positives and the remaining are negatives.
        masks
            Tensors containing the skiplist masks assocaited with each sentence feature (anchor, positives, negatives).
        with_backward
            Whether to compute the backward pass or not.
        """
        # We first cat them chunk-wise for anchor, positives, negatives
        embeddings_anchor = torch.cat(reps[0])  # (bsz, hdim)
        embeddings_other = [
            torch.cat([chunk_embed for chunk_embed in r]) for r in reps[1:]
        ]  # [(nneg * bsz, hdim)]

        batch_size = len(embeddings_anchor)
        labels = torch.tensor(
            range(batch_size), dtype=torch.long, device=reps[0][0].device
        )  # (bsz, (1 + nneg) * bsz)  Example a[i] should match with b[i]
        # Possibly gather the embeddings across devices to have more in-batch negatives. For GradCache, we only need to gather them to compute the scores matrix and nowhere else.
        # Note that we only gather the documents embeddings and not the queries embeddings, but are keeping gradients. This is to lower the memory usage, see https://github.com/mlfoundations/open_clip/issues/616
        if self.gather_across_devices:
            embeddings_other = [
                torch.cat(all_gather_with_gradients(embeddings))
                for embeddings in embeddings_other
            ]
            # Masks [0] is the anchor mask so we do not need to gather it (even though we are not using it for now anyways)
            # Also, we do gather without gradients for the masks as we do not backpropagate through them
            masks = [
                masks[0],
                *[torch.cat(all_gather(mask)) for mask in masks[1:]],
            ]
            rank = get_rank()
            # Adjust the labels to match the gathered embeddings positions
            labels = labels + rank * batch_size
        losses: list[torch.Tensor] = []
        do_query_expansion = (
            self.model.do_query_expansion
            if hasattr(self.model, "do_query_expansion")
            else self.model.module.do_query_expansion
        )
        for begin in tqdm.trange(
            0,
            batch_size,
            self.mini_batch_size,
            desc="Preparing caches",
            disable=not self.show_progress_bar,
        ):
            end = begin + self.mini_batch_size
            # We chunk the scores computation to avoid OOM because MaxSim can get expensive with large batch sizes/long documents
            scores = torch.cat(
                [
                    torch.cat(
                        [
                            self.score_metric(
                                embeddings_anchor[begin:end],
                                group_embeddings[
                                    g_start : min(
                                        g_start + self.mini_batch_size,
                                        len(group_embeddings),
                                    )
                                ],
                                queries_mask=masks[0][begin:end]
                                if not do_query_expansion
                                else None,
                                documents_mask=documents_mask[
                                    g_start : min(
                                        g_start + self.mini_batch_size,
                                        len(group_embeddings),
                                    )
                                ],
                            )
                            for g_start in range(
                                0, len(group_embeddings), self.mini_batch_size
                            )
                        ],
                        dim=1,
                    )
                    for group_embeddings, documents_mask in zip(
                        embeddings_other, masks[1:]
                    )
                ],
                dim=1,
            )
            # We don't want to average the loss across the mini-batch as mini-batch sizes can vary, which would create an issue similar to this one: https://huggingface.co/blog/gradient_accumulation#where-does-it-stem-from
            loss_mbatch = F.cross_entropy(
                input=scores / self.temperature,
                target=labels[begin:end],
                reduction="sum",
            )
            # Scale by world size when gathering across device
            if self.gather_across_devices:
                loss_mbatch *= get_world_size()

            if with_backward:
                loss_mbatch.backward()
                loss_mbatch = loss_mbatch.detach()
            losses.append(loss_mbatch)

        loss = sum(losses)
        if self.size_average:
            loss /= batch_size

        return loss

    def forward(
        self,
        sentence_features: Iterable[dict[str, Tensor]],
        labels: Optional[Tensor] = None,
    ) -> Tensor:
        """Compute the CachedConstrastive loss.

        Parameters
        ----------
        sentence_features
            List of tokenized sentences. The first sentence is the anchor and the rest are the positive and negative examples.
        labels
            The labels for the contrastive loss. Not used in this implementation, but kept for compatibility with Trainer.
        """
        # Step (1): A quick embedding step without gradients/computation graphs to get all the embeddings
        reps = []
        self.random_states = []  # Copy random states to guarantee exact reproduction of the embeddings during the second forward pass, i.e. step (3)
        # handle the model being wrapped in (D)DP and so require to access module first
        skiplist = (
            self.model.skiplist
            if hasattr(self.model, "skiplist")
            else self.model.module.skiplist
        )
        masks = extract_skiplist_mask(
            sentence_features=sentence_features, skiplist=skiplist
        )
        for sentence_feature in sentence_features:
            reps_mbs = []
            random_state_mbs = []
            for reps_mb, random_state in self.embed_minibatch_iter(
                sentence_feature=sentence_feature,
                with_grad=False,
                copy_random_state=True,
            ):
                reps_mbs.append(reps_mb.detach().requires_grad_())
                random_state_mbs.append(random_state)
            reps.append(reps_mbs)
            self.random_states.append(random_state_mbs)
        if torch.is_grad_enabled():
            # Step (2): Calculate the loss, backward up to the embeddings and cache the gradients wrt. to the embeddings
            loss = self.calculate_loss_and_cache_gradients(reps, masks)

            # Step (3): A 2nd embedding step with gradients/computation graphs and connect the cached gradients into the backward chain
            loss.register_hook(
                partial(
                    _backward_hook, sentence_features=sentence_features, loss_obj=self
                )
            )
        else:
            # If grad is not enabled (e.g. in evaluation), then we don't have to worry about the gradients or backward hook
            loss = self.calculate_loss(reps, masks)

        return loss

    @property
    def citation(self) -> str:
        return """
@misc{gao2021scaling,
    title={Scaling Deep Contrastive Learning Batch Size under Memory Limited Setup},
    author={Luyu Gao and Yunyi Zhang and Jiawei Han and Jamie Callan},
    year={2021},
    eprint={2101.06983},
    archivePrefix={arXiv},
    primaryClass={cs.LG}
}
"""
